# -*- python -*-
# $RCSfile: worker.py,v $
# $Revision: 1.89 $
# $Author: reida $
# $Date: 2007/03/16 17:39:57 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


## A worker is a wrapped generalized callback (from a menuitem) that will
## run in a separate thread. This callback is monitored through progress
## bars, and gtk progress bars.
##

from oof2.SWIG.common.IO import progressbar
from oof2.SWIG.common import threadstate
from oof2.SWIG.common import lock
from oof2.SWIG.common import ooferror
from oof2.common import debug
from oof2.common import mainthread
##from oof2.common import utils
from oof2.common.IO import reporter
from oof2.common.IO import threadmanager
from oof2.common import parallel_enable

if parallel_enable.enabled():
    from oof2.common.IO.socket2me import pipeToSocket

import sys
import threading
import time

# Worker classes.  The basic highest-level interface is that workers
# must have a "start" routine, which is called from the menu item
# after the worker is created, and it must also have a
# "getProgressBar" routine which takes a progress bar type.  The
# second one is required because the type of progress bar required is
# not generally known until run-time within the callback.

# Threading is handled with an extra layer of indirection -- the
# "ThreadedWorker" and subclasses are actually wrappers, which contain
# a "ThreadedWorkerCore" object.  These wrapper objects call their
# core object's "start" routine, and then manage the main thread,
# which may want to do complicated things before (or instead of)
# returning control to the caller.  The TextThreadedWorker draws the
# progress bar for the command.  The GUIThreadedWorkerBlock object,
# used in GUI mode, blocks until the just-launched thread terminates.
# This is useful when running scripts, for instance, when each script
# menu item (threadable or otherwise) runs on a subthread, but must
# not run concurrently with previous script commands.  This technique
# allows the GUI to be responsive and a script-reading progress bar to
# be presented during the read.

# Parallelization is fairly straightforward at this level -- menu
# items which are parallelized require a worker which is a subclass of
# ParallelWork, which itself only ensures that all the back-end
# machines receive the same menu item.

# Flag to indicate how to handle exceptions.  If set to 0, workers
# should catch exceptions and send them to the reporter.  If nonzero,
# workers should allow exceptions to propagate to the caller.  In the
# case of threaded workers, the core should propagate the exception
# data to the calling worker, which should re-raise it.
propagate_exceptions=0

# Worker is the the base class for the workers that are invoked by the
# menu items.  Subclasses either actually call the menu item's
# callback, or invoke a WorkerCore that calls the callback.  In the
# first case, the Worker and the WorkerCore may be the same object.
class Worker:
    def __init__(self, menuitem, args, kwargs):
        pass
    # Subclasses must redefine start().
    def start(self):
        pass

# WorkerCore is a base class for the part of the Worker that actually
# runs menu commands.  WorkerCore always runs on the same thread as
# the menu command.

class WorkerCore:
    def __init__(self, menuitem, args, kwargs):
        self.progressbar = None
        self.menuitem = menuitem
        self.args = args
        self.kwargs = kwargs
        self.bar_name = menuitem.bar_name
    def __getattr__(self, attrname):    # Stunt double for the menuitem
        return getattr(self.menuitem, attrname)    
    def get_name(self):
        return self.bar_name
    def getProgressBar(self, type):
        # Returns a progress bar object, with no representation.
        if self.progressbar is not None:
            return self.progressbar
        self.progressbar = progressbar.ProgressObject(type, self.bar_name)
        self.progressbar.registerWorker(self)
        return self.progressbar
    def initialize(self):
        pass
    def finalize(self):
        pass
    


class NonThreadedWorker(Worker, WorkerCore):
    def __init__(self, menuitem, args, kwargs):
        WorkerCore.__init__(self, menuitem, args, kwargs)

    # "Unthreadable" really means "Run on the same thread as the GUI,
    # and block until finished".  This code actually puts the callback
    # on the GUI thread no matter what.  If threading is not enabled,
    # then the mainthread.mainthread() will return 1, and even if it
    # doesn't, in that case mainthread.runBlock just runs it on the
    # calling thread, which in the unthreaded case is also the main
    # thread. 
    def start(self):
        if self.menuitem.callback is not None:
            if mainthread.mainthread(): # already on mainthread, just call
                func = self.menuitem.callback
                func_args = (self,) + self.args
                func_kwargs = self.kwargs
            else:               # use mainthread.runBlock to get to main thread
                func = mainthread.runBlock
                func_args = (self.menuitem.callback,
                             (self,)+self.args,
                             self.kwargs)
                func_kwargs = {}
                
            try:
                self.initialize()
                try:
                    self.menuitem.precall()
                    func(*func_args, **func_kwargs)
                    self.menuitem.postcall()
                except SystemExit:
                    raise
                # TODO SWIG1.3: After conversion to SWIG 1.3, OOF
                # exceptions will probably be subclasses of Exception.
                except (Exception, ooferror.ErrErrorPtr), exception:
                    if propagate_exceptions:
                        raise
                    else:
                        reporter.error(exception)
            finally:
                self.finalize()
                
################
    
class ThreadedWorkerCore(threading.Thread, WorkerCore):
    def __init__(self, menuitem, args, kwargs):
        WorkerCore.__init__(self, menuitem, args, kwargs)
        self.exception_data=None
        self.done = False
        self.threadstate = None
        threading.Thread.__init__(self)
        
    def run(self):
        cpbar = progressbar.getProgress()
        self.threadstate = cpbar.get_threadstate()
        try:
            self.initialize()
            try:
                if self.menuitem.callback is not None:
                    self.menuitem.precall()
                    self.menuitem.callback(self, *self.args, **self.kwargs)
                    self.menuitem.postcall()
            except SystemExit:
                raise
            # TODO SWIG1.3: After conversion to SWIG 1.3, OOF
            # exceptions will probably be subclasses of Exception.
            except (Exception, ooferror.ErrErrorPtr), exception:
                cpbar.set_failure()
                if propagate_exceptions:
                    self.exception_data=sys.exc_info()
                    sys.exc_clear()
                else:
                    reporter.error(exception)
        finally:
            self.finalize()
            self.done = True
            cpbar.make_defunct()
##            self.cleanupProgressBar()
            self.progressbar = None
            threadmanager.threadManager.removeWorker(self)

        

    def cancel(self):
        if self.threadstate:
            threadstate.cancelThread(self.threadstate)
        else:
            print "No threadstate available"


class ThreadedWorker(Worker):
    def __init__(self, menuitem, args, kwargs):
        self.worker = ThreadedWorkerCore(menuitem, args, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
        self.menuitem = menuitem        # for debugging
    def start(self):
        threadstate.cleanList()
        self.worker.start()
        self.meanwhile()                # back in the main thread...

    def meanwhile(self):  # Default, should be overridden if nontrivial
        pass              # behavior is needed.
    
    def join(self):                     # waits for thread to finish
        self.worker.join()


class TextThreadedWorker(ThreadedWorker):
    def __init__(self, menuitem, args, kwargs):
        ThreadedWorker.__init__(self, menuitem, args, kwargs)
        self.initialized = False
        
    def meanwhile(self):
        ## This function should only be called if in text mode and
        ## on the main thread.  It checks to see if the thread is
        ## finished and redraws the progress bar.  It sleeps a bit
        ## after each check so that it doesn't eat up cpu.  The sleep
        ## time is initially short, but grows (up to a point) on each
        ## iteration.
        naptime = 0.001                 # initial length of sleep time
        maxnap = 0.3                    # maximum length of sleep time
        napfactor = 2.0                 # amount to increase on each iteration
        prgrsbar = None                 # progressbar representation
        try:
            thread_running = True
            while thread_running:
                # self.worker.progressbar is volatile, so make a copy...
                pbar = self.worker.progressbar
                if pbar is not None:
                    self.progressbar = pbar
                    if not self.initialized:
                        prgrsbar = progressbar.set_ProgressBarRepresentation(
                            pbar)
                        self.initialized = True 
                    _draw_progressbar(prgrsbar)
                    if (pbar.query_stop() or pbar.get_success()):
                        thread_running = 0
                        _draw_progressbar(prgrsbar)
                else:
                    ## if progress bar never-ever set then exit loop
                    ## when worker finishes
                    if self.worker.done:
                        thread_running = False
                if thread_running:
                    time.sleep(naptime)
                    if naptime < maxnap:
                        naptime *= napfactor
            
        except KeyboardInterrupt:
            if prgrsbar is not None:
                prgrsbar.stop_it() ## asks for the thread to stop
                _draw_progressbar(prgrsbar)
            else:
                ## If there's no progress bar, there's no way to
                ## interrupt the thread...
                pass 

        # Wait for threads to finish before OOF-ing some more.  Text
        # mode doesn't support concurrent operations.
        self.join()

        # Check if the core has caught an exception and placed it in
        # exception_data.  Re-raise it so it gets passed through to
        # the caller of the menu item.
        if propagate_exceptions:
            if self.worker.exception_data is not None:
                raise self.worker.exception_data[0], \
                      self.worker.exception_data[1], \
                      self.worker.exception_data[2]
            

###############################

def _draw_progressbar(pbar):
    reporter.progressbar(pbar.buildprogressbar())
        

###############################


mpiLock = lock.Lock()

## this lock is necessary because MPI does not
## support more than one subthread at a
## time, without running the risk of dead-locking. In order
## to avoid this potential deadlocking
## problem, every parallel worker
## acquires a lock and releases it upon
## completion of the assigned task.
## the MPI lock will not be necessary
## anymore when a fully threadable
## implementation of MPI becomes
## available (MPI-2?).

##################################


# ParallelWorker classes are not instanced unless you're in parallel
# mode, so they are free to use socket2me in this file.

class ParallelThreadedWorkerCore(ThreadedWorkerCore):
    def __init__(self, menuitem, args, kwargs):
        ThreadedWorkerCore.__init__(self, menuitem, args, kwargs)
    def initialize(self):
        global mpiLock
        mpiLock.acquire()
    def finalize(self):
        global mpiLock
        mpiLock.release()



class ParallelWork:
    ## ParallelWork encapsulates all the start-up
    ## and clean-up functions that the ParallelWorker
    ## subclasses do, such as communicating between
    ## the different machines behind the scenes
    ## through the adequate (socket) machinery
    def __init__(self, menuitem, args, kwargs):
        pipeToSocket(menuitem, args, kwargs)

        ## front-end tells back-end to execute
        ## same menuitem. self.callback() sorts what
        ## is the task that should be performed
        ## by each processor.
        ## back-end bypasses this function

        
class ParallelWorker(NonThreadedWorker, ParallelWork):
    ## Unthreaded Parallel Worker
    def __init__(self,  menuitem, args, kwargs):
        NonThreadedWorker.__init__(self, menuitem, args, kwargs)
        ParallelWork.__init__(self, menuitem, args, kwargs)
        
    
        
class ParallelThreadedWorker (ThreadedWorker, ParallelWork):
    ## Threaded Parallel Worker
    def __init__(self,  menuitem, args, kwargs):
        self.worker = ParallelThreadedWorkerCore(menuitem, args, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
        ParallelWork.__init__(self, menuitem, args, kwargs)
        
    def meanwhile(self):
        pass ## back-end should always pass and give control
    ## to socketinput, after launching the job
    
class ParallelTextThreadedWorker(TextThreadedWorker, ParallelWork):
    ## Text mode Threaded Parallel Worker
    ## sould only be called by front end
    def __init__(self,  menuitem, args, kwargs):
        ## TODO: This __init__ doesn't call
        ## TextThreadedWorker.__init__, and in fact it does things
        ## differently than that class does.  This seems to indicate
        ## that the class hierarchy is incorrect. Probably there
        ## should be a common base class for
        ## ParallelTextThreadedWorker and TextThreadedWorker.  That
        ## base class's __init__ would just set self.initialized and
        ## do nothing else.
        self.worker = ParallelThreadedWorkerCore(menuitem, args, kwargs)
        threadmanager.threadManager.newWorker(self.worker)
        self.initialized = False
        ParallelWork.__init__(self, menuitem, args, kwargs)
        
def getThreadedParallelWorker(menuitem, argtuple, argdict):
    ## parallel workers
    try:
        from oof2.SWIG.common import mpitools
        if mpitools.Rank() == 0:
            ## front-end runs threaded workers with text progress bars
            return ParallelTextThreadedWorker(menuitem, argtuple, argdict)
        ## back-end runs a threaded worker too, but no text progress bar
        ## is available. Back-end, instead, runs a thread, and the
        ## "front-end' of the back-end waits on the socket parser
        return ParallelThreadedWorker(menuitem, argtuple, argdict)
    except ImportError:
            raise


###############################
# getThreadedWorker() is redefined if the GUI is loaded in workerGUI.py

def getThreadedWorker(menuitem, argtuple, argdict):
    ## serial workers
    return TextThreadedWorker(menuitem, argtuple, argdict)


