# -*- python -*-
# $RCSfile: adaptivemeshrefinement.py,v $
# $Revision: 1.23 $
# $Author: reida $
# $Date: 2007/05/07 15:59:57 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.common.IO import progressbar
from oof2.common import debug
from oof2.common import labeltree
from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import errorestimator
from oof2.engine import meshmod
from oof2.engine import rationalize
from oof2.engine import refine
from oof2.engine import skeletonboundary
from oof2.engine import skeletonelement
from oof2.engine import skeletonmodifier
from oof2.engine import subproblemcontext

SkeletonEdgeBoundary = skeletonboundary.SkeletonEdgeBoundary
ExteriorSkeletonEdgeBoundary = skeletonboundary.ExteriorSkeletonEdgeBoundary
SkeletonPointBoundary = skeletonboundary.SkeletonPointBoundary
ExteriorSkeletonPointBoundary = skeletonboundary.ExteriorSkeletonPointBoundary

# AdaptiveMeshRefinement
class AdaptiveMeshRefinement(meshmod.MeshModification):
    def __init__(self, estimator, criterion, degree, alpha,
                 rationalize=0):
        self.estimator = estimator  # ErrorEstimator instance
        self.criterion = criterion  # RefinementCriterion instance
        self.degree = degree  # RefinementDegree instance
        self.alpha = alpha
        self.rationalize = rationalize
        
    def doTheBoundary(self, old_skel, new_skel):
        # Point boundary propagation.
        for pbdy in old_skel.sheriffSkeleton().pointboundaries.values():
            pbdy.makeContextBoundary(
                None, pbdy.get_name(), old_skel).map(old_skel, new_skel)
        # Edge boundary propagation.
        for ebdy in old_skel.sheriffSkeleton().edgeboundaries.values():
            ebdy.makeContextBoundary(
                None, ebdy.get_name(), old_skel).map(old_skel, new_skel)
        
    def apply(self, meshcontext):
        try:
            defaultsubp = meshcontext.get_default_subproblem()
        except KeyError:
            raise ooferror.ErrUserError(
             "Adaptive Mesh Refinement requires a Subproblem named 'default'")
        pBar = progressbar.getProgress()

        # refine the skeleton.
        refinstance = refine.Refine(
            errorestimator.AdaptiveMeshRefine(meshcontext.path(),
                                              self.estimator),
            self.criterion, self.degree, self.alpha)

        
        old_skel = meshcontext.getSkeleton()
        new_skel = refinstance.applyAMR(old_skel)
        
        # Boundary propagation.
        self.doTheBoundary(old_skel, new_skel)
        # Only the second Skeleton need to be disconnected.
        if meshcontext.skeleton_buffer.isFirstInLine(old_skel):
            new_skel.disconnect()

        if pBar.query_stop():
            pBar.set_failure()
            pBar.set_message("Failed")
            
            return None
        else:
            pBar.set_progress(0.5)
            pBar.set_message("Mesh Refined / Post-processing.")
        
        # rationalize the skeleton
        if self.rationalize:
            newer_skel = new_skel.properCopy(fresh=True)
            targets = skeletonmodifier.AllElements()
            criterion = skeletonmodifier.AverageEnergy(0.8)  # alpha
            for ratreg in rationalize.Rationalizer.registry:
                # Create a Rationalizer from a Registration
                # with default values.
                ratmethod = ratreg()
                ratmethod(newer_skel, None, targets, criterion,
                          ratmethod.findAndFix)
            newer_skel.cleanUp()
            self.doTheBoundary(new_skel, newer_skel)
            new_skel = newer_skel

        meshcontext.pushSkeleton(new_skel)

        # convert the refined skeleton into an femesh
        old_femesh = meshcontext.femesh()
        # realmaterial is a function which tells the elements of the
        # created femesh how to get their materials.
        realmaterial = skeletonelement.SkeletonElement.realmaterial
        new_femesh = new_skel.femesh(meshcontext.elementdict,
                                     realmaterial)
        new_femesh.set_rwlock(meshcontext.rwLock)

        
        # This code is vaguely similar to that in copyMesh in
        # meshmenu.py (the callback for OOF.Mesh.Copy), but is *not*
        # identical.  That's because in this case we're replacing the
        # FEMesh inside a Mesh context, instead of simply creating a
        # whole new Mesh context.

        # Copy subproblems, their fields and equations.  The tricky
        # bit here is that the new subproblems must live in the same
        # Mesh as the old ones and must have the same names, but they
        # refer to different FEMeshes.  If the old subproblems are
        # deleted before the new ones are created, then their Field
        # values in the old FEMesh will be deleted before they can be
        # copied to the new FEMesh.  On the other hand, if they old
        # subproblems are kept, then their names will conflict with
        # the new ones.  Therefore, we have to create the new
        # subproblems with altered names, and change the names after
        # the old subproblems are deleted.  Also, we can't actually
        # insert the new subproblems into the Mesh context until the
        # new FEMesh has been installed...

        # First, gather data about subproblems and their fields and
        # equations.
        class SubProbState:
            pass
        subpstates = {}
        for old_subpctxt in meshcontext.subproblems():
            old_subp = old_subpctxt.getObject()
            subpstate = subpstates[old_subpctxt.path()] = SubProbState()
            subpstate.old_subp = old_subp
            subpstate.new_subp = old_subp.clone()
            subpstate.new_subp.set_time(old_subp.current_time())
            subpstate.fields = old_subp.all_fields() # all defined fields
            subpstate.activefields = [f for f in subpstate.fields
                                      if old_subp.is_active_field(f)]
            subpstate.eqns = old_subp.all_equations() # all active equations
            
        # replace old_femesh
        meshcontext.setFemesh(new_femesh)

        # Install new subproblems and their fields and equations
        for subppath, subpstate in subpstates.items():
            tempname = subproblemcontext.subproblems.uniqueName(subppath)
            subpstate.tempname = tempname
            meshcontext.newSubProblem(subpstate.new_subp,
                                      meshcontext.path() + ":" + tempname)
            for field in subpstate.fields:
                subpstate.new_subp.define_field(field)
                new_femesh.set_in_plane(field, old_femesh.in_plane(field))
            for field in subpstate.activefields:
                subpstate.new_subp.activate_field(field)
            for equation in subpstate.eqns:
                subpstate.new_subp.activate_equation(equation)

        # Copy Field values from the old FEMesh
        for field in meshcontext.all_subproblem_fields():
            new_femesh.init_field(old_skel, old_femesh, field)

        # Destroy old subproblems (this has to happen *after* fields
        # are copied) and give new subproblems their correct names.
        meshcontext.pause_writing()
        try:
            for subppath, subpstate in subpstates.items():
                oldname = labeltree.makePath(subppath)[-1]
                subpctxt = subproblemcontext.subproblems[subppath]
                subpctxt.begin_writing()
                try:
                    subpctxt.destroy()
                finally:
                    subpctxt.end_writing()
                temppath = meshcontext.path() + ':' + subpstate.tempname
                newctxt = subproblemcontext.subproblems[temppath]
                newctxt.begin_writing()
                try:
                    newctxt.rename(oldname)
                finally:
                    newctxt.end_writing()
                ## Passing the default subproblem to the FEMesh is a hack,
                ## intended to be removed once adaptive mesh refinement is
                ## moved from FEMesh to CSubProblem.
                if oldname == 'default':
                    new_femesh.setDefaultSubProblem(newctxt.getObject())
                    
        finally:
            meshcontext.resume_writing()
            
            

##        # copy field status
##        for field in meshcontext.all_fields():
##            new_femesh.define_field(field)
##            new_femesh.activate_field(field)
##            new_femesh.set_in_plane(field, old_femesh.in_plane(field))
        
##        # copy equation status
##        for equation in meshcontext.all_equations():
##            new_femesh.get_subproblem().activate_equation(equation)


        # put BC in new_femesh
        for (name, bc) in meshcontext.allBoundaryConds():
            bc.boundary_obj=new_femesh.getBoundary(bc.boundary)
            new_femesh.boundaries[bc.boundary].addCondition(bc)
        meshcontext.pause_writing()
        try:
            for subproblem in meshcontext.subproblems():
                subproblem.autoenableBCs()
        finally:
            meshcontext.resume_writing()
            
        old_femesh.destroy()
                
        # time stamp update
        meshcontext.changed()

        if pBar.query_stop():
            pBar.set_failure()
            pBar.set_message("Failed")
            return None
        else:
            pBar.set_progress(1.)
            pBar.set_message("Done.")

## TODO: Figure out where to put this.  It either needs to be on a
## page with a SubProblem WhoWidget, or needs to have a SubProblem
## WhoParameter itself.
            
registeredclass.Registration(
    'Adaptive Mesh Refinement',
    meshmod.MeshModification,
    AdaptiveMeshRefinement,
    ordering=4,
    params=[
    parameter.RegisteredParameter('estimator',
                                  errorestimator.ErrorEstimator,
                                  tip="Which error estimator to use."),
    parameter.RegisteredParameter('criterion',
                                  refine.RefinementCriterion,
                                  tip="Only consider elements meeting the criterion."),
    parameter.RegisteredParameter('degree',
                                  refine.RefinementDegree,
                                  tip="How to subdivide elements."),
    skeletonmodifier.alphaParameter,
    parameter.BooleanParameter('rationalize', 0, default=0,
                               tip='Rationalize mesh?')
    ],
    tip="Adaptively refine mesh based on error estimation.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/adaptive.xml')
    )
