// -*- C++ -*-
// $RCSfile: element.C,v $
// $Revision: 1.71 $
// $Author: langer $
// $Date: 2007/04/27 15:58:53 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#include "common/ooferror.h"
#include "common/swiglib.h"
#include "common/tostring.h"
#include "common/trace.h"
#include "engine/csubproblem.h"
#include "engine/edge.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/femesh.h"
#include "engine/flux.h"
#include "engine/masterelement.h"
#include "engine/material.h"
#include "engine/nodalequation.h"
#include "engine/node.h"
#include "engine/ooferror.h"
#include "engine/outputval.h"
#include "engine/shapefunction.h"
#include "engine/cskeleton.h"
#include <string>
#include <vector>

// ElementData constructors.
ElementData::ElementData(const std::string &nm) 
  : name_(nm)
{}  

Element::Element(PyObject *skelel, const MasterElement &me,
		 const std::vector<Node*> *nl, const Material *mat)
  : master(me),
    nodelist(*nl),
    matl(mat),
    explicitMatl(false),
    exterior_edges(0),
    skeleton_element(skelel),
    cskeleton_element(0),
    edgeset(0)
{

//   Trace("Element::Element " + me.name());
//   int nn = me.nnodes();
//   for(int i=0; i<nn; i++) {
//     nodelist[i] = (*nl)[i]; 
//   }
  acquirePyLock();
  SWIG_GetPtrObj(skeleton_element, (void **)&cskeleton_element, "_CSkeletonElement_p");
  Py_XINCREF(skeleton_element);
  releasePyLock();
}

Element::~Element() {
  if(edgeset != 0) {
    for(int i=0; i<2*master.nedges(); i++)
      delete edgeset[i];
    delete [] edgeset;
  }
  delete exterior_edges;
  acquirePyLock();
  Py_XDECREF(skeleton_element);
  releasePyLock();
}

std::string Element::repr() const {
  std::string rep =  master.name();
  for(int i=0; i<nnodes(); i++)
    rep += " " + to_string(*nodelist[i]);
  return rep;
}
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
// Index stuff.

void Element::set_index(int i) {
  index_=i;
}

const int &Element::get_index() const {
  return index_;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

int Element::nnodes() const { return master.nnodes(); }
int Element::nmapnodes() const { return master.nmapnodes(); }
int Element::nfuncnodes() const { return master.nfuncnodes(); }
int Element::ncorners() const { return master.ncorners(); }
int Element::nexteriorfuncnodes() const { return master.nexteriorfuncnodes(); }

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// refreshMaterial() is called when Materials have been reassigned in a
// Microstructure after the FEMesh has been created.  The Elements
// need to ask their SkeletonElements for the new Material.

void Element::refreshMaterial(PyObject *skeleton) {
  if(!explicitMatl) {
    acquirePyLock();
    try {
      // Call skeletonelement.material(skeleton)
      PyObject *pymat = PyObject_CallMethod(skeleton_element, "material", "O",
					    skeleton);
      if(!pymat) {
	pythonErrorRelay();
      }
      if(pymat == Py_None) {
	matl = 0;
      }
      else {
	// Extract the C++ Material* from the Python object.
	SWIG_GetPtrObj(pymat, (void**)(&matl), "_Material_p");
	Py_XDECREF(pymat);
      }
    }
    catch (...) {
      releasePyLock();
      throw;
    }
    releasePyLock();
  }
}

void Element::setExplicitMaterial(const Material *m) {
  matl = m;
  explicitMatl = true;
}

const Material *Element::getExplicitMaterial() const {
  if(explicitMatl)
    return matl;
  return 0;
}

void Element::removeExplicitMaterial(PyObject *skeleton) {
  explicitMatl = false;
  refreshMaterial(skeleton);	// gets Material from Microstructure
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

void Element::make_linear_system(CSubProblem *subproblem, 
				 LinearizedSystem &system) const {
//   Trace("Element::make_stiffness for element " + to_string(index_));
  const Material *mat = material();
  if(mat) {
    mat->begin_element(subproblem, this);
    mat->make_linear_system(subproblem, this, system);
    mat->end_element(subproblem, this);
  }
}

// Post-processing, which is after equilibration.
void Element::post_process(CSubProblem *subproblem) const {
  const Material *mat = material();
  if(mat) {
    mat->post_process(subproblem, this);
  }
}

// void Element::begin_material_computation(FEMesh *mesh) const {
//   const Material *mat = material();
//   if(mat)
//     mat->begin_element(mesh, this);
// }

// void Element::end_material_computation(FEMesh *mesh) const {
//   const Material *mat = material();
//   if(mat)
//     mat->end_element(mesh, this);
// }

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

int Element::appendData(ElementData *x) {
  el_data.push_back(x);
  return el_data.size()-1;
}

void Element::setData(int i, ElementData *x) {
  el_data[i]=x;
}

ElementData *Element::getData(int i) {
  return el_data[i];
}

// This call does retrieval-by-name, and so should be used
// as infrequently as can be managed.
ElementData *Element::getDataByName(const std::string &name) {
  int i=getIndexByName(name);
  if (i>=0)
    return getData(i);
  else
    return 0;
}

// Overwrites the data indexed under the indicated name, appends
// it, whichever.
void Element::setDataByName(ElementData *x) {
  int i=getIndexByName(x->name());
  if (i>=0)
    // NB Does this overwrite have implications for reference counts
    // of the referred-to Python data?  
    setData(i,x);
  else 
    appendData(x);
}


// Retrieval by name should be done infrequently, to avoid the search.
int Element::getIndexByName(const std::string &searchname) {
  for(std::vector<ElementData*>::size_type i=0; i<el_data.size(); i++) {
    if (el_data[i]->name() == searchname) {
      return i;
    }
  }
  // Not found -- this is permissible, return something.
  return -1;
}




//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ElementNodeIterator Element::node_iterator() const {
  return ElementNodeIterator(*this);
}

ElementMapNodeIterator Element::mapnode_iterator() const {
  return ElementMapNodeIterator(*this);
}

ElementFuncNodeIterator Element::funcnode_iterator() const {
  return ElementFuncNodeIterator(*this);
}

ElementCornerNodeIterator Element::cornernode_iterator() const {
  return ElementCornerNodeIterator(*this);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// conversion between real coordinates and master element coordinates

Coord Element::from_master(const MasterPosition &mc) const {
  Coord p(0., 0.);
  for(ElementMapNodeIterator n=mapnode_iterator(); !n.end(); ++n) {
    p += n.shapefunction(mc) * n.node()->position();
  }
  return p;
}

static const double tolerancesq = 1.e-10; // should be settable by user
static const int maxiter = 100;	// should also be settable by user

MasterCoord Element::to_master(const Coord &x) const {
  // Use Newton's method to solve from_master(xi) - x = 0 for xi.
  // xi -> xi + J^-1 * (x - from_master(xi))
  // This code is wrong in three dimensions
  MasterCoord xi(0, 0);
  Coord dx = x - from_master(xi);
  int iter = 0;
  do {				// Newton iteration
    double jac[spacedim][spacedim]; // use simple matrix repr. for 2x2
    for(SpaceIndex i=0; i<spacedim; ++i)
      for(SpaceIndex j=0; j<spacedim; ++j)
	jac[i][j] = jacobian(i, j, xi);
    
    double dj = jac[0][0]*jac[1][1] - jac[0][1]*jac[1][0];
    dj = 1./dj;
    // xi --> xi + J^1*(x - from_master(xi))
    xi(0) += ( jac[1][1]*dx(0) - jac[0][1]*dx(1))*dj;
    xi(1) += (-jac[1][0]*dx(0) + jac[0][0]*dx(1))*dj;
    dx = x - from_master(xi);	// reevaluate rhs
  } while((norm2(dx) > tolerancesq) && (++iter < maxiter));
  return xi;
}

// Superconvergent patch recovery
int Element::nSCpoints() const {
  return master.nSCpoints();
}

const MasterCoord &Element::getMasterSCpoint(int i) const {
  return master.getSCpoint(i);
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

double Element::Jdmasterdx(SpaceIndex i, SpaceIndex j, const GaussPoint &g)
  const
{
  // This is wrong in three dimensions, since it returns the
  // components of the inverse of the 2x2 Jacobian matrix:
  //
  //         | J11  -J01 |
  //  J^-1 = |           | / |J|
  //         |-J10   J00 |

  double sum = 0;
  if(i == j) {
    int ii = 1 - i;
    for(ElementMapNodeIterator ni=mapnode_iterator(); !ni.end(); ++ni) {
      sum += (ni.node()->position())(ii) * ni.masterderiv(ii, g);
    }
  }
  else {			// i != j
    for(ElementMapNodeIterator ni=mapnode_iterator(); !ni.end(); ++ni)
      sum -= (ni.node()->position())(i) * ni.masterderiv(j, g);
  }
  return sum;
}

double Element::Jdmasterdx(SpaceIndex i, SpaceIndex j, const MasterCoord &mc)
  const
{
  // This also is wrong in three dimensions.

  double sum = 0;
  if(i == j) {
    int ii = 1 - i;
    for(ElementMapNodeIterator ni=mapnode_iterator(); !ni.end(); ++ni)
      sum += (ni.node()->position())(ii) * ni.masterderiv(ii, mc);
  }
  else {			// i != j
    for(ElementMapNodeIterator ni=mapnode_iterator(); !ni.end(); ++ni)
      sum -= (ni.node()->position())(i) * ni.masterderiv(j, mc);
  }
  return sum;
}


double Element::Jdmasterdx(SpaceIndex i, SpaceIndex j, const Coord &coord)
  const
{
  return Jdmasterdx(i, j, to_master(coord));
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// J(i,j) = d(real_coord i)/d(master_coord j)

double Element::jacobian(SpaceIndex i, SpaceIndex j, const GaussPoint &g) const
{
  double jac = 0.0;
  for(ElementMapNodeIterator ni=mapnode_iterator(); !ni.end(); ++ni)
    jac += (ni.node()->position())(i) * ni.masterderiv(j, g);

  return jac;
}

double Element::jacobian(SpaceIndex i, SpaceIndex j, const MasterCoord &mc)
  const
{
  double jac = 0.0;
  for(ElementMapNodeIterator ni=mapnode_iterator(); !ni.end(); ++ni)
    jac += (ni.node()->position())(i) * ni.masterderiv(j, mc);
  return jac;
}

double Element::det_jacobian(const GaussPoint &g) const {
  return master.mapfunction->det_jacobian(this, g);
}

double Element::det_jacobian(const MasterCoord &mc) const {
  return master.mapfunction->det_jacobian(this, mc);
}

int Element::shapefun_degree() const {
  return master.shapefunction->degree();
}

int Element::dshapefun_degree() const {
  return master.shapefunction->deriv_degree();
}

int Element::mapfun_degree() const {
  return master.mapfunction->degree();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

OutputValue Element::outputField(const Field &field, const MasterPosition &pos)
  const
{
  OutputValue val = field.newOutputValue();
  for(ElementFuncNodeIterator node=funcnode_iterator(); !node.end(); ++node) {
    double sfvalue = node.shapefunction(pos);
    val += sfvalue*field.output(node);
  }
  return val;
}

std::vector<OutputValue> *
Element::outputFields(const Field &field,
		      const std::vector<MasterCoord*> *coords) const
{
  std::vector<OutputValue> *results = new std::vector<OutputValue>;
  results->reserve(coords->size());
  for(std::vector<MasterCoord*>::size_type i=0; i<coords->size(); i++) {
    OutputValue val = field.newOutputValue();
    for(ElementFuncNodeIterator node=funcnode_iterator(); !node.end(); ++node) {
      double sfvalue = node.shapefunction(*(*coords)[i]);
      // Field::output returns an appropriate type of 0 if the Field
      // isn't defined on the Node.
      val += sfvalue*field.output(node);
    }
    results->push_back(val);
  }
  return results;
}

// std::vector<OutputValue> *
// Element::outputFieldsAnyway(const CSubProblem *mesh, const Field &field,
// 			    const std::vector<MasterCoord*> *coords) const
// {
//   // Return zero if the field isn't defined.  Don't throw an exception.
//   if(mesh->is_defined_field(field)) 
//     return outputFields(mesh, field, coords);
//   // Field isn't defined.
//   std::vector<OutputValue> *results = new std::vector<OutputValue>;
//   results->reserve(coords->size());
//   for(std::vector<MasterCoord*>::size_type i=0; i<coords->size(); i++) {
//     results->push_back(field.newOutputValue());
//   }
//   return results;
// }

// This version takes a vector of MasterCoords, not MasterCoord*s, and
// is used by Edge::outputFields.
std::vector<OutputValue> *
Element::outputFields(const Field &field,
		      const std::vector<MasterCoord> &coords) const
{
  std::vector<OutputValue> *results = new std::vector<OutputValue>;
  results->reserve(coords.size());
  for(std::vector<MasterCoord>::size_type i=0; i<coords.size(); i++) {
    OutputValue val = field.newOutputValue();
    for(ElementFuncNodeIterator node=funcnode_iterator(); !node.end(); ++node) {
      double sfvalue = node.shapefunction(coords[i]);
      // Field::output returns an appropriate type of 0 if the Field
      // isn't defined on the Node.
      val += sfvalue*field.output(node);
    }
    results->push_back(val);
  }
  return results;
}

// std::vector<OutputValue> *
// Element::outputFieldsAnyway(const CSubProblem *mesh, const Field &field,
// 		      const std::vector<MasterCoord> &coords) const
// {
//   if(mesh->is_defined_field(field))
//     return outputFields(mesh, field, coords);
//   std::vector<OutputValue> *results = new std::vector<OutputValue>;
//   results->reserve(coords.size());
//   for(std::vector<MasterCoord>::size_type i=0; i<coords.size(); i++) {
//     results->push_back(field.newOutputValue());
//   }
//   return results;
// }

std::vector<OutputValue> *
Element::outputFieldDerivs(const Field &field, SpaceIndex *deriv_component,
			   const std::vector<MasterCoord*> *coords) const
{
  std::vector<OutputValue> *results = new std::vector<OutputValue>;
  results->reserve(coords->size());
  for(std::vector<MasterCoord*>::size_type i=0; i<coords->size(); i++) {
    OutputValue val = field.newOutputValue();
    for(ElementFuncNodeIterator node=funcnode_iterator(); !node.end(); ++node) {
      double dsfvalue = node.dshapefunction(*deriv_component, *(*coords)[i]);
      val += dsfvalue*field.output(node);
    }
    results->push_back(val);
  }
  return results;
}

OutputValue Element::outputFieldDeriv(const Field &field,
				      SpaceIndex *deriv_component,
				      const MasterPosition &pos) const
{
  OutputValue val = field.newOutputValue();
  for(ElementFuncNodeIterator node=funcnode_iterator(); !node.end(); ++node) {
    double dsfvalue = node.dshapefunction(*deriv_component, pos);
    val += dsfvalue*field.output(node);
    }
  return val;
}

OutputValue Element::outputFlux(const FEMesh *mesh, const Flux &flux,
				const MasterPosition &pos) const
{
  return flux.output(mesh, this, pos);
}

std::vector<OutputValue> *
Element::outputFluxes(const FEMesh *mesh, const Flux &flux,
		      const std::vector<MasterCoord*> *coords) const
{
  std::vector<OutputValue> *results = new std::vector<OutputValue>;
  results->reserve(coords->size());
  for(std::vector<MasterCoord*>::size_type i=0; i<coords->size(); i++)
    results->push_back(flux.output(mesh, this, *(*coords)[i]));
  return results;
}

// Create a vector of the values of the Element's degrees of freedom,
// in canonical order.  Called by Flux::evaluate and
// Flux::evaluate_flux.

VECTOR_D Element::localDoFs() const {
  VECTOR_D doflist(ndof(), 0.0);
  for(std::vector<Field*>::size_type fi=0; fi<Field::all().size(); fi++) {
    Field &field = *Field::all()[fi];
    // loop over field components
    for(IteratorP fcomp=field.iterator(ALL_INDICES); !fcomp.end(); ++fcomp) {
      // loop over nodes	
      for(ElementFuncNodeIterator node=funcnode_iterator(); !node.end(); ++node)
	{
	  if(node.hasField(field)) {
	    DegreeOfFreedom *dof = field(node, fcomp.integer());
	    doflist[node.localindex(field, fcomp)] = dof->value();
	  }
	}
    }
  }
  return doflist;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// The main edge function. 

BoundaryEdge *Element::newBndyEdge(const FuncNode *n0, const FuncNode *n1)
  const
{
//   Trace("Element::newBndyEdge " + to_string(*n0) + " " + to_string(*n1));
  // find ElementCornerNodeIterators for each FuncNode
  ElementCornerNodeIterator alpha = this->cornernode_iterator();
  // It may not be sufficient to compare pointers to Nodes, since the
  // Nodes can have different derived types.  Use
  // operator==(Node&,Node&) and operator!=(Node&,Node&) instead.
  while((*alpha.node() != *n0) && !alpha.end()) {
    ++alpha;
  }
  if(alpha.end()) {
    throw ErrUserError("Element::newBndyEdge: Nodes not corners of element");
  }
  bool forward;
  ElementCornerNodeIterator beta = alpha + 1;
  if(*beta.node() == *n1) {
    // second node is in front of first
    forward = true;
  }
  else {			// second node isn't in front
    beta = alpha + (ncorners() - 1);
    if(*beta.node() != *n1) {
      // it's not behind either
      throw ErrUserError("Element::newBndyEdge: Inconsistent node set");
    }
    forward = false;		// it was behind
  }

  // find out which edge of the masterelement contains alpha and beta
  const MasterEdge *medge = masterelement().masteredge(alpha, beta);

  // put the endpoints and all intermediate FuncNodes in the BoundaryEdge
  BoundaryEdge *ed = new BoundaryEdge(this, medge->func_size());
  int increment;
  if(forward)
    increment = 1;
  else
    increment = nexteriorfuncnodes() - 1;

  ed->add_node(alpha.funcnode_iterator());
  ElementExteriorNodeIterator middle = alpha.exteriornode_iterator()+increment;
  while(*middle.node() != *beta.node()) {
    ed->add_node(middle);
    middle += increment;
  }
  ed->add_node(beta.funcnode_iterator());

  return ed;
}

BoundaryEdge *Element::getBndyEdge(const FuncNode *n0, const FuncNode *n1) {
  BoundaryEdge *ed;

//   Trace("Element::getBndyEdge " + to_string(*n0) + " " + to_string(*n1));
  // find_edge will correctly detect nullness of the edgelist.
  if( (ed=find_b_edge(n0,n1))==0 ) { 
    ed=newBndyEdge(n0,n1);
    if(ed!=0) {
      add_b_edge(ed);
      return ed;
    }
    else 
      throw ErrUserError("Element::getBndyEdge: Unable to make edge.");
    // Probable cause is that the input nodes are not adjacent
    // corners in the derived element type.
  }
  return ed; // Otherwise return the non-null result from find_edge.
}   

BoundaryEdge *Element::find_b_edge(const FuncNode *n0, const FuncNode *n1) {

  if( edgeset==0 )  // There are no nodes on the edge list, therefore
    return 0;       // the requested node is not on the edge list.

  for(int i=0;i<master.nedges();i++) {
    if(edgeset[i]==0)   // Run away at the first occurence of zero.
      return 0; 
    if( edgeset[i]->edge_match(n0,n1))
      return edgeset[i];
  }

  return 0;  // If you got all the way to the end, then it wasn't there.
}

// The edge list is not allocated unless needed, and it needs to
// be 2x the number of geometric edges, because edges are 
// directed.
void Element::add_b_edge(BoundaryEdge *ed_in) {
 
  if(edgeset==0) {  // If the edge-list is null, create it.
    edgeset=new BoundaryEdge*[2*master.nedges()]; //del in destructor.
    for(int i=0;i<2*master.nedges();i++) // Edges are directed.
      edgeset[i]=0;
    edgeset[0]=ed_in;  // List was null, so ours is the first.
    return;
  }
  
  int i=0;  // Otherwise, find the right index for insertion.
  while( (edgeset[i]!=0)&&(i<master.nedges()) )
    i++;

  if(i!=master.nedges())	// If this conditional is false, you
    edgeset[i]=ed_in;		// have gone off the end.  Throw exception.

  return;
}

std::vector<Edge*> *Element::perimeter() const {
  std::vector<Edge*> *brim = new std::vector<Edge*>(master.nedges());
  for(ElementCornerNodeIterator it=cornernode_iterator(); !it.end(); ++it) {
    (*brim)[it.index()] = new Edge(this,
				   it.mastercoord(), (it+1).mastercoord());
  }
  return brim;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

std::vector<GaussPoint*>* Element::integration_points(int order) const {
  std::vector<GaussPoint*>* r = new std::vector<GaussPoint*>;
  for(GaussPointIterator g = integrator(order); !g.end(); ++g) {
    r->push_back(g.gausspointptr());
  }
  return r;
}
  

GaussPointIterator Element::integrator(int order) const {
  return GaussPointIterator(this, (order < 0 ? 0 : order));
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

int Element::ndof() const {
  int n = 0;
  for(ElementFuncNodeIterator ni=funcnode_iterator(); !ni.end(); ++ni)
    n += ni.funcnode()->ndof();
  return n;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Generic area calculation, done by integrating 1.

double Element::area() const {
  double a = 0.0;
  for(GaussPointIterator gpt = integrator(0); !gpt.end(); ++gpt) {
    a += gpt.gausspoint().weight();
  }
  return a;
}

MasterCoord Element::center() const {
  return master.center();
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

bool Element::exterior(const MasterCoord &a, const MasterCoord &b) const {
  // Are points a and b on the same exterior edge?
  if(!exterior_edges)
    return 0;
  // Let the master do the actual computation, so as to avoid roundoff errors.
  return master.exterior(a, b, *exterior_edges);
}

void Element::set_exterior(const Node &n0, const Node &n1) {
  if(!exterior_edges)
    exterior_edges = new std::vector<ElementCornerNodeIterator>;
  // find which edge these nodes are on.
  for(ElementCornerNodeIterator it=cornernode_iterator(); !it.end(); ++it) {
    const Node *node0 = it.node();
    const Node *node1 = (it+1).node();
    if((node0==&n0 && node1==&n1) || (node1==&n0 && node0==&n1)) {
      exterior_edges->push_back(it);
//       cerr << "Element::set_exterior it=" << it << endl;
//       cerr << "    n0=" << n0 <<     "    n1=" << n1 << endl;
//       cerr << " node0=" << *node0 << " node1=" << *node1 << endl;
      return;
    }
  }
}

void Element::dump_exterior() const {  // debugging
  if(!exterior_edges) return;
  std::cerr << "Element::dump_exterior " << *this << std::endl;
  for(std::vector<ElementCornerNodeIterator>::size_type i=0;
      i<exterior_edges->size(); i++)
    std::cerr << "   " << (*exterior_edges)[i] << std::endl;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

std::ostream &operator<<(std::ostream &os, const Element &el) {
  os << el.master.name();
  for(ElementCornerNodeIterator it=el.cornernode_iterator(); !it.end(); ++it)
    os << " " << it.node()->position();
  return os;
}
