# -*- python -*-
# $RCSfile: mesh.py,v $
# $Revision: 1.178 $
# $Author: reida $
# $Date: 2007/05/16 20:42:15 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

# The Mesh class is the Who class wrapper for the C++ FEMesh class,
# associating a Skeleton and miscellaneous bookkeeping information
# with it.  Mesh itself is a Who class, which means it can participate
# in the layer-display process.

from oof2.SWIG.common import lock
from oof2.SWIG.common import ooferror
from oof2.SWIG.common import switchboard
from oof2.SWIG.common import timestamp
from oof2.SWIG.engine import masterelement
from oof2.SWIG.engine import material
from oof2.SWIG.engine import entiremeshsubproblem
from oof2.common import debug
from oof2.common import enum
from oof2.common import ringbuffer
from oof2.common import utils
from oof2.common.IO import parameter
from oof2.common.IO import whoville
from oof2.common import parallel_enable
import oof2.SWIG.engine.field

import copy
import sets
import string
from math import *

# Helper class for keeping track of the cross-sections -- remembers
# what the most-recently-added one was by remembering the order.
class CrossSectionSet:
    def __init__(self):
        self.data = {}
        self.namelist = []
        self._selection = None
        self.timestamp = timestamp.TimeStamp()
    def new(self, name, cs):
        self.data[name]=cs
        self.namelist.append(name)
        self.timestamp.increment()
        self._selection = name           # always select new cross section
        switchboard.notify("cross sections changed")
    def remove(self, name):
        del self.data[name]
        self.namelist.remove(name)
        self.timestamp.increment()
        if self._selection == name:
            if self.namelist:
                self._selection = self.namelist[-1]
            else:
                self._selection = None
        switchboard.notify("cross sections changed")
    def all_names(self):
        return self.namelist
    def __getitem__(self, key):
        return self.data[key]
    def selected(self):
        try:
            return self.data[self._selection]
        except KeyError:
            return None
    def selectedName(self):
        return self._selection
    def select(self, name):
        self._selection = name
        self.timestamp.increment()
        switchboard.notify("cross sections changed")
    def deselect(self):
        self._selection = None
        self.timestamp.increment()
        switchboard.notify("cross sections changed")
    def rename(self, oldname, newname):
        cs = self.data[oldname]
        newname = utils.uniqueName(newname, self.namelist, exclude=oldname)
        self.namelist[self.namelist.index(oldname)] = newname
        del self.data[oldname]
        self.data[newname] = cs
        if oldname == self._selection:
            self._selection = newname
        switchboard.notify("cross sections changed")
        return newname
    def replace(self, name, cross_section):
        self.data[name] = cross_section
        self.timestamp.increment()
        switchboard.notify("cross sections changed")
        
class StatusItem:
    def __init__(self, state=0, message=""):
        self.state = state
        self.msg = message
    def __eq__(self, val):
        return self.state == val
    def __repr__(self):
        return self.msg

######################

class FEMeshMapOrder(enum.EnumClass(
    ('1', 'linear map functions'),
    ('2', 'quadratic map functions'))):
    tip = "Order of mapping functions."
    discussion = """
    <para>Order of mapping functions.
    </para>"""                    

class FEMeshShapeOrder(enum.EnumClass(
    ('1', 'linear shape functions'),
    ('2', 'quadratic shape functions'))):
    tip = "Order of shape functions."
    discussion = """
    <para>Order of shape functions.
    </para>"""                    

######################

skel_stack_size = 50
# TODO CLEANUP: RingBuffer is not a terribly good choice for this.
class SkeletonBuffer(ringbuffer.RingBuffer):
    def makeCurrent(self, index):  # make the given index the current position.
        self.currentpos = index
        return self.current()

    def isFirstInLine(self, skel):
        return self.data.index(skel)==0
    
class Mesh(whoville.Who):
    def __init__(self, name, classname, femesh, parent,
                 skeleton=None, elementdict=None,
                 materialfactory=None):
        whoville.Who.__init__(self, name, classname, femesh, parent)
        # Share the mesh-level lock with the contained femesh.
        femesh.set_rwlock(self.rwLock)
        # bdy bdy.
        self.bdyconditions = {}
        # Skeletons will be stored in the ring buffer.  Skeleton
        # objects are stored, not SkeletonContexts.
        self.skeleton_buffer = SkeletonBuffer(skel_stack_size)
        self.pushSkeleton(skeleton)
        # This will add skeleton to the buffer and so some more
        
        self.elementdict = elementdict
        self.materialfactory = materialfactory
        femesh.set_parent_mesh(self)
        # Cross-section stuff
        self.cross_sections = CrossSectionSet()

        # Timestamps.  Boundarieschanged is incremented whenever a
        # boundary or boundary condition is added or removed.
        # Materialschanged is incremented for material operations.
        self.boundarieschanged = timestamp.TimeStamp() # last bdy change
        self.materialschanged = timestamp.TimeStamp()
        self.fieldsinitialized = timestamp.TimeStamp()

        self.initializers = {}          # Field initializers

        # Cached info about which materials are used.  Caches are
        # recomputed when refreshMaterials is called.
        self.materiallist = None          # list of all Materials used
        self.expmaterialdict = None       # explicitly added Materials

        # Point boundaries used by periodic boundary conditions
        self.periodicPointBoundaries = {}

    def createDefaultSubProblem(self):
        subp = entiremeshsubproblem.entiremeshreg()
        self.newSubProblem(subp,
            self.path() + ":" + defaultSubProblemName)
        ## Passing the default subproblem to the FEMesh is a hack,
        ## intended to be removed once adaptive mesh refinement is
        ## moved from FEMesh to CSubProblem.
        self.getObject().setDefaultSubProblem(subp)

    def pushSkeleton(self, skeleton):
        self.skeleton_buffer.push(skeleton)
        self.setSkeleton(skeleton)  # set the current skeleton
        skeleton.add_mesh(self)

    # getTimeStamp generically has a gfxwindow arg so that it can work
    # with WhoPoxies.  But Mesh.getOwnTimeStamp is called by
    # SubProblemContext.status, which doesn't have a gfxwindow, so the
    # arg is optional here.
    def getTimeStamp(self, gfxwindow):
        subtimes = [sub.getTimeStamp(gfxwindow) for sub in self.subproblems()]
        # 'max(self.getOwnTimeStamp(gfxwindow), *subtimes)' is a
        # syntax error if subtimes is an empty list, so we have to
        # check.
        if subtimes:
            return max(self.getOwnTimeStamp(gfxwindow), *subtimes)
        return self.getOwnTimeStamp(gfxwindow)
    def getOwnTimeStamp(self, gfxwindow=None):
        return max(self.materialschanged, self.boundarieschanged,
                   self.fieldsinitialized)

    def femesh(self):
        return self.getObject()

    def size(self):
        return self.skeleton.size()

    # utility function from "femesh"
    def nelements(self):
        return self.getObject().nelements()

    def nnodes(self):
        return self.getObject().nnodes()
    
    # Required for compatibility with Display objects which work for
    # both Mesh and Skeleton who classes.  
    def mesh(self):
        return self.getObject()

    def getSkeleton(self):  # returns the current skeleton object (not context)
        return self.skeleton

    def setSkeleton(self, skeleton):
        self.skeleton = skeleton

    def setFemesh(self, femesh):
        self._obj = femesh  # replace the old one with the new one
        self._obj.set_parent_mesh(self)

    def getMicrostructure(self):
        return self.getParent().getParent().getObject()

    def getMasterElementType(self, nsides):
        try:
            return self.elementdict[nsides].name()
        except KeyError:
            return '---'

    def destroy(self):
        global meshes
        self.skeleton.remove_mesh(self)
        del self.skeleton               # may not be necessary but doesn't hurt
        self.getObject().destroy()
        # Remove from the enclosing WhoClass.  This nulls out parent's "._obj".
        meshes.remove(self.path())

    def lockAndDelete(self):
        self.reserve()
        try:
            # Delete subproblems.  Since some subproblems depend on
            # other subproblems, deleting one may trigger the deletion
            # of others.  Therefore we can't simply loop over
            # self.subproblems() with a "for" loop.
            while len(self.subproblems()) > 0:
                subproblem = self.subproblems()[0]
                subproblem.begin_writing()
                subproblem.destroy()
                subproblem.end_writing()
            self.begin_writing()
            try:
                self.destroy()
            finally:
                self.end_writing()
        finally:
            self.cancel_reservation()

    def newSubProblem(self, subproblem, path):
        # "subproblem" is a SubProblem object.  Adding it to the
        # SubProblem Who class creates a SubProblemContext.
        from oof2.engine import subproblemcontext  # avoid import loop
        return subproblemcontext.subproblems.add(path, subproblem, self)
    
    def subproblemNames(self):
        from oof2.engine import subproblemcontext
        path = self.path()
        if not path:
            raise ooferror.ErrPyProgrammingError("No path to mesh '%s'!?"
                                                 % self.name())
        subpaths = subproblemcontext.subproblems.keys(base=path)
        ## subpaths is a list of relative paths of length 1.
        return [s[0] for s in subpaths]

    def subproblems(self):
        from oof2.engine import subproblemcontext
        path = self.path()
        return [subproblemcontext.subproblems[path + ':' + s]
                for s in self.subproblemNames()]
        
    # No equations are ever activated directly in a mesh, but the
    # boundary conditions need to know which equations are available.
    # So, this routine polls the subproblems and returns the set of
    # available equations.
    def all_subproblem_equations(self):
        eqns = utils.OrderedSet()
        for s in self.subproblems():
            for e in s.all_equations():
                eqns.add(e)
        return list(eqns)
    def all_subproblem_equations_bc(self): # just eqns that can have bc's.
        eqns = utils.OrderedSet()
        for s in self.subproblems():
            for e in s.all_equations_bc():
                eqns.add(e)
        return list(eqns)


    # Analysis can be done on any compound Field from any SubProblem.
    def all_compound_subproblem_fields(self):
        fields = utils.OrderedSet()
        for s in self.subproblems():
            for f in s.all_compound_fields():
                fields.add(f)
        return list(fields)


    # Really all fields, including out-of-plane parts.
    def all_subproblem_fields(self):
        fields = utils.OrderedSet()
        for s in self.subproblems():
            for f in s.all_fields():
                fields.add(f)
        return list(fields)
    
    
    def all_active_subproblem_fields(self):
        fields = utils.OrderedSet()
        for s in self.subproblems():
            for f in s.all_compound_fields():
                if s.is_active_field(f):
                    fields.add(f)
        return list(fields)

    # Or any Flux
    def all_subproblem_fluxes(self):
        fluxes = utils.OrderedSet()
        for s in self.subproblems():
            for f in s.all_fluxes():
                fluxes.add(f)
        return list(fluxes)

    def is_defined_field(self, field):
        for s in self.subproblems():
            if s.is_defined_field(field):
                return True
        return False

    def set_in_plane_field(self, field, inplane):
        self.getObject().set_in_plane(field, inplane)
        ## Activate the out-of-plane field in all subproblems only if
        ## this field is active and not in-plane.
        zfield = field.out_of_plane()
        for s in self.subproblems():
            subp = s.getObject()
            if inplane:
                subp.deactivate_field(zfield)
            else:
                if subp.is_active_field(field):
                    subp.activate_field(zfield)

    def precompute_all_subproblems(self):
        for s in self.subproblems():
            s.getObject().precompute()

    def precompute_all_materials(self):
        materials = material.getMaterials(self.getMicrostructure())
        for mat in materials:
            mat.precompute_all_properties(self.getObject())
                
    def allCrossSectionNames(self):
        return self.cross_sections.all_names()

    def addCrossSection(self, name, cs):
        self.cross_sections.new(name, cs)

    def removeCrossSection(self, name):
        self.cross_sections.remove(name)

    def uniqueCSName(self, name):
        return utils.uniqueName(name, self.cross_sections.all_names())

    def selectedCS(self):
        return self.cross_sections.selected()

    def selectedCSName(self):
        return self.cross_sections.selectedName()

    def getCrossSection(self, name):
        return self.cross_sections[name]

    def selectCrossSection(self, name):
        self.cross_sections.select(name)

    def deselectCrossSection(self):
        self.cross_sections.deselect()

    def renameCrossSection(self, oldname, newname):
        actualnewname = self.cross_sections.rename(oldname, newname)
        ## This switchboard call is here, rather than in
        ## CrossSectionSet.rename, because it needs to send the Mesh
        ## as an argument.
        switchboard.notify(('cross section renamed', self),
                           oldname, actualnewname)

    def replaceCrossSection(self, name, cross_section):
        self.cross_sections.replace(name, cross_section)
        
    # Boundary retrieval functions -- the interface needs to know
    # the names.

    def getBoundary(self, name):
        return self.getObject().boundaries[name]

    # Names of nontrivial boundaries -- does not return a complete
    # list, only returns the names of those objects with size > 0.
    # This is the meaning of the "Finite" modifier.
    def boundaryNames(self):
        names= self.getObject().getFiniteBoundaryNames()
        return names

    def pointBoundaryNames(self):
        if parallel_enable.enabled():
            # RCL: This gives the list of point boundaries, including
            # boundaries with zero size (no nodes).  Do the same for
            # edge boundaries. We do this so that all the conventional
            # boundaries (left,right,bottom,top,
            # topleft,topright,bottomleft,bottomright) show up in the
            # BoundaryCondition (sub)pages (for process 0).  No need
            # to do this for boundaryNames (just above) because it
            # doesn't seem to be used.
            return self.getObject().pointbdynames
        else:
            return self.getObject().getFinitePointBdyNames()

    def visiblePointBoundaryNames(self):
        if parallel_enable.enabled():
            names = self.getObject().pointbdynames
        else:
            names = self.getObject().getFinitePointBdyNames()

        # make copy of original list because we will alter
        # it in the loop
        names_orig = names[:]
        for name in names_orig:
            if not self.getObject().boundaries[name].visible:
                names.remove(name)

        return names

    def edgeBoundaryNames(self):
        if parallel_enable.enabled():
            return self.getObject().edgebdynames
        else:
            return self.getObject().getFiniteEdgeBdyNames()

    def periodicEdgeBoundaryNames(self):
        if parallel_enable.enabled():
            names = self.getObject().edgebdynames
        else:
            names = self.getObject().getFiniteEdgeBdyNames()

        # Periodic boundary conditions apply to two mesh boundaries
        periodic_names = []
        for i in xrange(len(names)):
            edgeInfo0 = self.getBoundary(names[i]).whichPeriodicEdge(self)
            if edgeInfo0:
                # search rest of list for the opposite boundary with opposite direction
                for j in xrange(i+1,len(names)):
                    edgeInfo1 = self.getBoundary(names[j]).whichPeriodicEdge(self)
                    if edgeInfo1 and edgeInfo0[0]==edgeInfo1[0] and \
                           edgeInfo0[1]!=edgeInfo1[1] and edgeInfo0[2]==-edgeInfo1[2]:
                        periodic_names.append(names[i]+" "+names[j])
        return periodic_names
    
    # Post-construction-time changes to the boundaries can also happen.
    
    def newEdgeBoundary(self, name, skeleton_bdy):
        realmesh = self.getObject()
        realbdy = realmesh.newEdgeBoundary(name)
        for e in skeleton_bdy.edges:
            skelel = e.segment.getElements()[0]
            realel = realmesh.getElement(skelel.meshindex)

            edge_nodes = e.get_nodes()
            realn0 = realmesh.getNode( edge_nodes[0].meshindex )
            realn1 = realmesh.getNode( edge_nodes[1].meshindex )
            realbdy.addEdge(realel.getBndyEdge(realn0,realn1))
        switchboard.notify("mesh boundaries changed", self)



    def newPointBoundary(self, name, skeleton_bdy):
        realmesh = self.getObject()
        realbndy = realmesh.newPointBoundary(name)
        for node in skeleton_bdy.nodes:
            realbndy.addNode(realmesh.getNode(node.meshindex))
        switchboard.notify("mesh boundaries changed", self)


    # Periodic boundary conditions use point boundaries containing
    # partnered nodes.  We only want to create the point boundaries once.
    def getPeriodicPointBoundaries(self, name):

        try:
            ppb = self.periodicPointBoundaries[name]
 
        except KeyError:

            boundaries = string.splitfields(name)

            try:
                boundary_obj0 = self.getBoundary(boundaries[0])
            except:
                raise ooferror.ErrSetupError(
                    "There is no boundary named '%s'!" % boundaries[0])
            try:
                boundary_obj1 = self.getBoundary(boundaries[1])
            except:
                raise ooferror.ErrSetupError(
                    "There is no boundary named '%s'!" % boundaries[1])       

            realmesh = self.getObject()
            nodes0 = boundary_obj0.edgeset.nodes()[:]
            nodes1 = boundary_obj1.edgeset.nodes()[:]
            nodes1.reverse()
            self.periodicPointBoundaries[name] = []
            for i in xrange(len(nodes0)):
                if fabs(nodes0[i].position().x - nodes1[i].position().x)<1e-14 or fabs(nodes0[i].position().y - nodes1[i].position().y)<1e-14:
                    newname = name + "_" + str(i)
                    bdy = realmesh.newPointBoundary(newname,visible=False)
                    bdy.addNode(nodes0[i])
                    bdy.addNode(nodes1[i])
                    self.periodicPointBoundaries[name].append(bdy)
                else:
                    raise ooferror.ErrSetupError(
                        "Nodes along periodic edge boundaries do not match up")

        return self.periodicPointBoundaries[name]
    
            
    def removeBoundary(self, name):
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==name:
                bc.remove_from_mesh()
                self.bdys_changed()
        self.getObject().removeBoundary(name)
        switchboard.notify("mesh boundaries changed", self)
        

    def renameBoundary(self, oldname, newname):
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==oldname:
                bc.boundary=newname
        self.getObject().renameBoundary(oldname, newname)
        switchboard.notify("mesh boundaries changed", self)

    def replacePointBoundary(self, name, new_skel_boundary):
        #del self.getObject().boundaries[name]
        self.getObject().removePointBoundary(name)
        self.newPointBoundary(name, new_skel_boundary)
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==name:
                bc.boundary_obj = self.getObject().boundaries[name]
                bc.boundary_obj.addCondition(bc)
                self.bdys_changed()
        switchboard.notify("mesh boundaries changed", self)
        

    def replaceEdgeBoundary(self, name, new_skel_boundary):
        #del self.getObject().boundaries[name]
        self.getObject().removeEdgeBoundary(name)
        self.newEdgeBoundary(name, new_skel_boundary)
        for (bcname, bc) in self.allBoundaryConds():
            if bc.boundary==name:
                bc.boundary_obj = self.getObject().boundaries[name]
                bc.boundary_obj.addCondition(bc)
                self.bdys_changed()
        self.boundarieschanged.increment()
        switchboard.notify("mesh boundaries changed", self)
    
    # Boundary condition manipulation functions.  The "promotion"
    # argument indicates whether or not this modification involves the
    # promotion of a Float BC to a Dirichlet BC.  In this case, no
    # real change is occurring in the mesh, and in particular, you
    # don't want to draw it any differently, so "self.changed" is not
    # called.  Recipients of the "boundary conditions changed" signal
    # also get the "promotion" argument, and should behave
    # appropriately.
    
    # As of 2/12/07 this has been superceded by the visible flag,
    # which is a member of every BC. The visible flag handles any
    # cases where we want to hide a BC in the GUI.
    def addBdyCondition(self, name, bc):
        try:
            self.bdyconditions[name]
        except KeyError:
            self.bdyconditions[name]=bc
            switchboard.notify("boundary conditions changed",
                               self, name, bc.isVisible())
        else:
            raise ooferror.ErrSetupError(
                "Duplicate boundary condition name %s." % name )
        if bc.isVisible():
            self.bdys_changed()

    def uniqueBCName(self, name):
        return utils.uniqueName(name, self.bdyconditions.keys())

    def renameBdyCondition(self, oldname, newname):
        try:
            bc = self.bdyconditions[oldname]
        except KeyError:
            raise ooferror.ErrSetupError("Can't find boundary condition %s!"
                                         % oldname)
        newname = utils.uniqueName(newname, self.allBndyCondNames(),
                                   exclude = oldname)
        if newname != oldname:
            del self.bdyconditions[oldname]
            self.bdyconditions[newname] = bc

        self.boundarieschanged.increment()
        switchboard.notify("boundary conditions changed",
                           self, newname, bc.isVisible())

    def rmBdyCondition(self, name):
        self.bdyconditions[name].disconnect()
        visible = self.bdyconditions[name].isVisible()
        del self.bdyconditions[name]
        switchboard.notify("boundary conditions changed",
                           self, name, visible)
        if visible:
            self.bdys_changed()
        
    def allBoundaryConds(self):
        return self.bdyconditions.items()

    def allBndyCondNames(self):
        return self.bdyconditions.keys()

    def getBdyCondition(self, name):
        return self.bdyconditions[name]
    
    # Make a copy of the boundary condition, but on the indicated
    # boundary.  Caller is responsible for adding it to the mesh,
    # because that operation should remain at the menu level.
    def copyBdyCondition(self, bc, boundary):
        return self.bdyconditions[bc].copy(boundary)

    def refreshMaterials(self):
        self.reserve()
        self.begin_writing()
        try:
            self.getObject().refreshMaterials(self.skeleton)
            self.makeMaterialLists()
            self.materialschanged.increment()
            for s in self.subproblems():
                s.getObject().requirePrecompute()
        finally:
            self.end_writing()
            self.cancel_reservation()
        switchboard.notify("mesh changed", self)

    def makeMaterialLists(self):
        ## TODO: Why is this called so often?
        implicitMats = sets.Set()
        self.expmaterialdict = {}
        for element in self.getObject().element_iterator():
            mat = element.getExplicitMaterial()
            if mat:
                try:
                    elist = self.expmaterialdict[mat]
                except KeyError:
                    self.expmaterialdict[mat] = elist = []
                elist.append(element.get_index())
            else:
                mat = element.material()
                if mat is not None:
                    implicitMats.add(mat)
        implicitMats.union_update(self.expmaterialdict.keys())
        self.materiallist = list(implicitMats)

    def explicitMaterialElements(self):
        # Returns a dictionary whose keys are Materials and whose
        # values are lists of Element indices for Elements to which
        # those Materials have been explicitly assigned.
        if self.expmaterialdict is None:
            self.makeMaterialLists()
        return self.expmaterialdict

    def explicitMaterials(self):
        # Returns a list of Materials explicitly assigned to Elements
        # in the Mesh (as opposed to those implicitly assigned via the
        # Microstructure).
        return self.explicitMaterialElements().keys()

    def allMaterials(self):
        if self.materiallist is None:
            self.makeMaterialLists()
        return self.materiallist

    def changed(self):
        for subproblem in self.subproblems():
            subproblem.changed()

    # For boundary/boundary-condition changes, which require
    # the solver to be re-run, but which do not require the stiffness
    # matrix to be rebuilt.
    def bdys_changed(self):
        self.boundarieschanged.increment()
        switchboard.notify("mesh changed", self)

    def initialize_field(self, field, initializer):
        self.initializers[field] = initializer
        initializer.apply(self.getObject(), field)
        self.fieldsinitialized.increment() # timestamp

    def remove_initializer(self, field):
        try:
            del self.initializers[field]
        except KeyError:
            pass

        
    
    def update_fields(self):
        # Field definitions have changed in a subproblem.
        # definedfields = self.all_compound_subproblem_fields()
        # all_allowed_inits = definedfields + [x.out_of_plane for x in
        #                                      definedfields]
        definedfields = self.all_subproblem_fields()
        for field, initializer in self.initializers.items():
            if field not in definedfields:
                self.remove_initializer(field)
        
    def get_initializer(self, field):
        try:
            return self.initializers[field]
        except KeyError:
            return None

    def enclosingElement(self, point):
        skel = self.getSkeleton()
        skelel = skel.enclosingElement(point)
        meshel = self.getObject().getElement(skelel.meshindex)
        return meshel

    def get_default_subproblem(self):
        return self.get_subproblem(defaultSubProblemName)

    def get_subproblem(self, name):
        from oof2.engine import subproblemcontext 
        return subproblemcontext.subproblems[self.path()+":"+name]

    def compare(self, other, tolerance):
        tol2 = tolerance*tolerance
        mymesh = self.getObject()
        othermesh = other.getObject()

        # compare node positions
        mynodes = mymesh.node_iterator()
        othernodes = othermesh.node_iterator()
        while not (mynodes.end() or othernodes.end()):
            mynode = mynodes.node()
            othernode = othernodes.node()
            if (mynode.position() - othernode.position())**2 > tol2:
                return "Node outside of tolerance, %s" % \
                       mynode.position() - othernode.position()
            mynodes.next()
            othernodes.next()
        if not (mynodes.end() and othernodes.end()):
            return "Wrong number of nodes"

        # compare subproblems
        mysubprobs = self.subproblems()
        othersubprobs = other.subproblems()
        if len(mysubprobs) != len(othersubprobs):
            return "Wrong number of subproblems"
        # subproblem order may be different.  Check names before
        # comparing subproblems.
        subpnames = [s.name() for s in mysubprobs]
        osubpnames = [s.name() for s in othersubprobs]
        subpnames.sort()
        osubpnames.sort()
        if subpnames != osubpnames:
            return "Subproblem names don't match: %s!=%s" % (subpnames,
                                                             osubpnames)
        for subpname in subpnames:
            subpctxt = self.get_subproblem(subpname)
            osubpctxt = other.get_subproblem(subpname)
            subp = subpctxt.getObject()
            osubp = osubpctxt.getObject()
            mesh = subpctxt.getParent().getObject()
            omesh = osubpctxt.getParent().getObject()
            # Check that the same Fields are defined on the subproblems
            fields = subpctxt.all_compound_fields()
            ofields = osubpctxt.all_compound_fields()
            fields.sort()
            ofields.sort()
            if fields != ofields:
                return "Fields don't match in subproblem " + subpname

            # Check that the subproblems have the same sets of nodes
            mynodeiter = subp.funcnode_iterator()
            othernodeiter = osubp.funcnode_iterator()
            if mynodeiter.size() != othernodeiter.size():
                return "Different numbers of nodes in subproblem %s" % subpname

            # Create a sorted list of nodes, because funcnode_iterator
            # doesn't return them in a guaranteed order
            mynodes = []
            othernodes = []
            while not mynodeiter.end():
                mynodes.append(mynodeiter.node())
                mynodeiter.next()
            while not othernodeiter.end():
                othernodes.append(othernodeiter.node())
                othernodeiter.next()
            mynodes.sort(_nodesorter)
            othernodes.sort(_nodesorter)
            
            for field in fields:
                # Check field state
                if subp.is_active_field(field) != osubp.is_active_field(field):
                    return ("Field activity differs for %s on subproblem %s"
                            % (`field`, subpname))
                if mesh.in_plane(field) != omesh.in_plane(field):
                    return ("Field planarity differs for %s on subproblem %s"
                            % (`field`, subpname))
                oop = mesh.in_plane(field) and field.out_of_plane()
                
                # Check that the field values agree at the nodes
                for mynode, othernode in zip(mynodes, othernodes):
                    if mynode.position() != othernode.position():
                        return ("Bad node positions for subproblem %s: %s!=%s"
                                % (subpname, mynode.position(),
                                   othernode.position()))
                    if _fielddiff(field, mynode, othernode) > tol2:
                        print (mynode,
                               [field.value(mynode, i)
                                for i in range(field.ndof())],
                               othernode, [field.value(othernode, i)
                                           for i in range(field.ndof())])
                        return ("%s values differ for subproblem %s" %
                                (`field`, subpname))
                    if oop and _fielddiff(oop, mynode, othernode) > tol2:
                        return ("Out-of-plane %s values differ for subproblem %s",
                                (`field`, subpname))
        return 0                        # Success
                        
# Utility functions used by Mesh.compare
def _fielddiff(field, nodeA, nodeB):
    fieldA = field.output(nodeA).valueClone().value_list()
    fieldB = field.output(nodeB).valueClone().value_list()
    diffs = [x-y for x,y in zip(fieldA, fieldB)]
    sqdiffs = reduce(lambda x,y: x+y*y, diffs, 0.0)/len(diffs)
    return sqdiffs
def _nodesorter(nodeA, nodeB):
    return cmp(nodeA.position(), nodeB.position())
                

defaultSubProblemName = "default"
        
from oof2.engine.skeletoncontext import skeletonContexts

meshes = whoville.WhoClass(
    'Mesh',
    ordering=300,
    parentClass=skeletonContexts,
    instanceClass=Mesh,
    proxyClasses=['<topmost>','<contourable>'])

utils.OOFdefine('meshes', meshes)
