# -*- python -*-
# $RCSfile: meshmod.py,v $
# $Revision: 1.15 $
# $Author: rlua $
# $Date: 2007/03/22 22:30:04 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# Mesh modification operations.  There shouldn't be too many of these,
# since most modifications are performed on a Skeleton instead.

from oof2.common import debug
from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import materialmanager
from oof2.engine import skeletoncontext
from oof2.engine.IO import materialmenu
from oof2.engine.IO import materialparameter
from oof2.engine.IO import skeletongroupparams

##############

class MeshModTargets(registeredclass.RegisteredClass):
    registry = []
    tip = "Target elements for Mesh modification."
    discussion = """<para>
    Objects in the <classname>MeshModTargets</classname> class are
    used as the <varname>target</varname> argument of <xref
    linkend='RegisteredClass:MeshModification'/> subclasses. They
    specify to which &mesh; elements a modification applies.
    </para>"""

class AllMeshElements(MeshModTargets):
    def __call__(self, meshcontext):
        els = []
        iter = meshcontext.getObject().element_iterator()
        while not iter.end():
            els.append(iter.element())
            iter.next()
        return els

registeredclass.Registration(
    'All Elements',
    MeshModTargets,
    AllMeshElements,
    ordering=0,
    tip="Operate on all Mesh elements",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/all_elements.xml')
    )
                             
class SelectedMeshElements(MeshModTargets):
    def __call__(self, meshcontext):
        mesh = meshcontext.getObject()
        skel = meshcontext.skeleton
##        skelcontext = skeletoncontext.skeletonContexts[skel.name()]
        skelcontext = skeletoncontext.skeletonContexts[
            skeletoncontext.extractSkeletonPath(meshcontext.path())]
        elements = skelcontext.elementselection.retrieveFromSkeleton(skel)
        return [mesh.getElement(el.meshindex) for el in elements]

registeredclass.Registration(
    'Selected Elements',
    MeshModTargets,
    SelectedMeshElements,
    ordering=1,
    tip="Operate on Mesh elements that are selected in the corresponding Skeleton.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/selected_elements.xml')
    )

class MeshElementGroup(MeshModTargets):
    def __init__(self, group):
        self.group = group
    def __call__(self, meshcontext):
        mesh = meshcontext.getObject()
        skel = meshcontext.skeleton
##        skelcontext = skeletoncontext.skeletonContexts[skel.name()]
        skelcontext = skeletoncontext.skeletonContexts[
            skeletoncontext.extractSkeletonPath(meshcontext.path())]
        elements = skelcontext.elementgroups.get_groupFromSkeleton(self.group,
                                                                   skel)
        return [mesh.getElement(el.meshindex) for el in elements]

registeredclass.Registration(
    'Element Group',
    MeshModTargets,
    MeshElementGroup,
    ordering=2,
    params=[
    skeletongroupparams.ElementGroupParameter('group',
                                              tip='Assign the Material to elements in this group.')
    ],
    tip="Operate on Mesh elements in a given element group.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/element_group.xml'))
                      

###############
    
class MeshModification(registeredclass.RegisteredClass):
    registry = []
    # Subclasses must define an "apply" function.
    # Also, if a Subclass needs a different type of progress bar,
    # it should be overridden.
    def get_progressbar_type(self):
        return "continuous"

    tip = "Tools to modify a Mesh."
    discussion = """<para>
    <classname>MeshModification</classname> objects modify &meshes;.
    They are used as the <varname>modifier</varname> argument of the
    <xref linkend='MenuItem:OOF.Mesh.Modify'/> command.
    </para>"""

class AssignMaterial(MeshModification):
    def __init__(self, material, target):
        self.material = material
        self.target = target
    def apply(self, meshcontext):
        matl = materialmanager.getMaterial(self.material)
        mesh = meshcontext.getObject()
        elements = self.target(meshcontext)
        for element in elements:
            oldmat = element.getExplicitMaterial()
            if oldmat:
                mesh.decrementExplicitMaterial(oldmat)
            mesh.incrementExplicitMaterial(matl)
            element.setExplicitMaterial(matl)
        meshcontext.materialschanged.increment()

registeredclass.Registration(
    'Assign Material',
    MeshModification,
    AssignMaterial,
    ordering=1,
    params=[
    materialparameter.MaterialParameter('material',
                                        tip='Material to assign to the elements.'),
    parameter.RegisteredParameter('target', MeshModTargets,
                                  tip='Which elements to assign the Material to.')
    ],
    tip="Assign a Material to elements.  This is *optional*, and overrides Material assignments from the Microstructure.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/assign_material.xml')
    )


class RemoveMaterial(MeshModification):
    def __init__(self, target):
        self.target = target
    def apply(self, meshcontext):
        skel = meshcontext.skeleton
        mesh = meshcontext.getObject()
        for element in self.target(meshcontext):
            oldmat = element.getExplicitMaterial()
            if oldmat:
                mesh.decrementExplicitMaterial(oldmat)
                element.removeExplicitMaterial(skel)
        meshcontext.materialschanged.increment()

registeredclass.Registration(
    'Remove Material',
    MeshModification,
    RemoveMaterial,
    ordering=2,
    params=[
    parameter.RegisteredParameter('target', MeshModTargets,
                                 tip="Remove the Material from these elements")
    ],
    tip="Remove a Material explicitly assigned to elements.  This *restores* the implicit Material assignments from the Microstructure.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/remove_material.xml')
    )

