// -*- C++ -*-
// $RCSfile: property.h,v $
// $Revision: 1.58 $
// $Author: langer $
// $Date: 2007/04/27 15:58:53 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifndef PROPERTY_H
#define PROPERTY_H

// forward declarations
class Property;

#include <Python.h>             // for PyObject declaration
#include "common/pythonexportable.h"
// #include "common/timestamp.h"
#include <string>
#include <vector>
#include <map>

// things defined elsewhere
class CSubProblem;
class EdgeSet;
class Element;
class ElementFuncNodeIterator;
class FEMesh;
class Field;
class Flux;
class FluxData;
class MasterPosition;
class Material;
class OutputVal;
class PropertyOutput;


// Property objects are universally created from PropertyRegistration
// objects' call methods, from Python.

// Since properties must be visible in Python, they must be SWIG'd,
// and in general should also have a "python part" of the SWIG'd
// interface, conventionally placed in an ".spy" file.  The python
// part of the interface must include registration of the property,
// i.e. the creation of a PropertyRegistration object with appropriate
// data.  The registration code must run at import-time.

// The property registration object must include the name of the
// property, the property's class, module, and a numerical "ordering"
// which is relative to all other property registration objects, and
// controls the presentation of this property in the GUI.  Following
// that, the property should indicate the parameters it requires, the
// fields it uses, and the fluxes to which it contributes.  It should
// indicate what type of property it is, and any outputs to which it
// contributes.

// The property type broadly indicates the role of this property in
// the material, and should be defined such that it only makes sense
// to have one property of each type in a material.  For instance, any
// property dependent on displacement and contributing to stress could
// advertise itself as being of type "Elasticity".  The OOF solver
// will then allow at most one property of type "Elasticity" per
// material.  The types are not predefined, property writers may make
// new types without modifying the engine code -- OOF will simply
// insist that at most one property of each defined type be present in
// each material.

class Property: virtual public PythonExportable<Property> {
private:
  const std::string name_;	// name of this instance
  std::string classname_;   // For PythonExportable-ability.
  std::string modulename_;  // Ditto
  std::vector<const Field*> fields_reqd; // fields reqd to compute this property
  Property(const Property&);	// prohibited
//   TimeStamp timestamp;
//   bool active_;			// see find_active and is_active below.
  typedef std::map<const CSubProblem*, bool> ActivityMap;
  ActivityMap activity;
public:
  Property(const std::string &nm, PyObject *registration);
  virtual ~Property();
  //
  // As a RegisteredCClass, Property must host a Python "registry"
  // entry, which is a list of Python objects.  This entry is
  // created as class data in the property.spy file.

  // This particular property's registration must also be stored, and
  // be retrievable.
  PyObject *registration_;
  PyObject *registration() const;

  const std::string &name() const { return name_; }
  // The following are required for a base class of PythonExportable.
  virtual const std::string &classname() const { return classname_; }
  virtual const std::string &modulename() const { return modulename_; }

//   const TimeStamp getTimeStamp() const { return timestamp; }

  // A Property is active if all the Fields it requires are defined on
  // a Mesh.  Activity is computed during the precomputation steps of
  // stiffness matrix construction and output computation.
  bool is_active(const CSubProblem*) const; // actually does the computation
  void find_active(const CSubProblem*); // computes and caches 
  bool currently_active(const CSubProblem*) const; // returns cached value
  
  // List of fields required to compute this Property
  const std::vector<const Field*> &fields_required() const {
    return fields_reqd;
  }
  // This field is required to compute this Property.  The derived
  // Properties don't call this directly, since they don't necessarily
  // know about compound Fields.  Instead, the Properties call
  // Field::registerProperty(Property*), which calls
  // Property::require_field().  This is done automatically in
  // Property::bookkeeping(), which is called by
  // Material::cross_reference().
  void require_field(const Field&);
  void bookkeeping(Material*);

  // The remaining virtual functions have default null bodies, so that
  // each Property only has to define the relevant ones:

  // Properties may need to find other properties in the material
  // (using Material::fetchProperty).  This function is called to tell
  // them that it's time to do that. It should throw an exception if
  // it's unsuccessful.
  virtual void cross_reference(Material*) {}

  // compute things that don't depend on Element
  virtual void precompute(FEMesh*) {}

  virtual bool constant_in_space() const = 0;

  // These routines allow Properties to precompute and store
  // mesh-specific data.  Since Materials and Properties are shared
  // between Meshes, Properties can't store mesh-specific data
  // themselves.  set_mesh_data stores data in the Mesh, and
  // get_mesh_data retrieves it. clear_mesh_data is a callback, called
  // when the Mesh is being destroyed and the data must be deleted.
  void set_mesh_data(FEMesh*, void *) const;
  void *get_mesh_data(const FEMesh*) const;
  virtual void clear_mesh_data(FEMesh*, void*) const {}

  // these functions are called when beginning and ending the
  // computations on an Element, allowing Element-dependent
  // precomputation and caching
  virtual void begin_element(const CSubProblem*, const Element*) {}
  virtual void end_element(const CSubProblem*, const Element*) {}


  // This function is called after equilibration, to allow the 
  // computation of auxiliary fields which may depend on equilibrium
  // fluxes -- the canoncal example is plasticity, where the
  // yield condition is stress-dependent.
  virtual void post_process(CSubProblem *, const Element *) const {}

  // this function returns the polynomial order (in x and y) of the
  // quantity computed by fluxmatrix.
  virtual int integration_order(const CSubProblem*, 
				const Element*) const;

  // Compute the matrix which, when multiplied by the vector of nodal
  // field values, gives the value of the specified Flux at the given
  // Position.  The rows of the matrix are indexed by the Flux
  // components; the columns are indexed by the local (element-wise)
  // dof indices of the Field components.  The matrix is stored inside
  // the Flux object -- see flux.h.
  virtual void fluxmatrix(const FEMesh*, const Element*, 
			  const ElementFuncNodeIterator&,
			  const Flux*, FluxData*, const MasterPosition&)
    const { return; }

  // The fluxrhs function works similarly to the fluxmatrix function,
  // but makes contributions to the rhs of global linear system,
  // either directly by way of the nodal equation values, or
  // indirectly by adding an offset to the flux itself.  It lies
  // outside of the ElementFuncNodeIterator loop.  Like the
  // fluxmatrix, the fluxrhs and fluxoffset are stored in the Flux
  // object.  See comments in flux.C for more info.
  virtual void fluxrhs(const FEMesh*, const Element*, const Flux*,
		       FluxData*, const MasterPosition&)
    const { return; }

  virtual void output(const FEMesh*, const Element*, const PropertyOutput*,
		      const MasterPosition&, OutputVal*)
    const { return; }

  virtual bool is_symmetric(const CSubProblem*) const;
};

std::ostream &operator<<(std::ostream &, const Property&);


class TimePropertyBase {
private:
  bool is_td;
public:
  TimePropertyBase();
  virtual ~TimePropertyBase();
  bool get_time_dependency() const {return is_td;}
};



class ScalarTimeDerivPrefactor : public Property, public TimePropertyBase {
protected:
  double generic_property_param; // TODO: Storing this here feels
				 // wrong.  There should probably be
				 // an abstract base class with a pure
				 // virtual fetch_value().  Then the
				 // value could be stored only where
				 // it's needed (ie, in a
				 // ConstantPrefactor class, or
				 // whatever).
public:
  ScalarTimeDerivPrefactor(const std::string &nm, PyObject *registration):
    Property(nm, registration), 
    TimePropertyBase(),
    generic_property_param(0)
  {}
  ScalarTimeDerivPrefactor(const std::string &nm, PyObject *registration, const double val):
    Property(nm, registration), 
    TimePropertyBase(),
    generic_property_param(val)
  {}
  virtual ~ScalarTimeDerivPrefactor() {}
  virtual double fetch_value() const; // default returns
				      // generic_property_param. Subclass
				      // may return arbitrary (scalar)
				      // functional form
};



#endif

