// -*- C++ -*-
// $RCSfile: permittivity.C,v $
// $Revision: 1.21 $
// $Author: langer $
// $Date: 2007/05/10 21:22:19 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

// heat conductivity

#include <oofconfig.h>
#include "common/coord.h"
#include "engine/celectricfield.h"
#include "engine/element.h"
#include "engine/elementnodeiterator.h"
#include "engine/field.h"
#include "engine/flux.h"
#include "engine/property/permittivity/permittivity.h"
#include "engine/fieldindex.h"
#include "engine/material.h"
#include "engine/property/orientation/orientation.h"
#include "common/space.h"
#include "engine/nodalequation.h"
#include "engine/IO/propertyoutput.h"

DielectricPermittivity::DielectricPermittivity(PyObject *registry, 
				   const std::string &nm)
  : Property(nm, registry)
{
  voltage = dynamic_cast<ScalarField*>(Field::getField("Voltage"));
  total_polarization = dynamic_cast<VectorFlux*>
    (Flux::getFlux("Total_Polarization"));
}

int DielectricPermittivity::integration_order(const CSubProblem *mesh,
					      const Element *el) const
{
  if(voltage->in_plane(mesh))
    return el->dshapefun_degree();
  return el->shapefun_degree();
}

void DielectricPermittivity::fluxmatrix(const FEMesh *mesh,
					const Element *el,
					const ElementFuncNodeIterator &j,
					const Flux *flux, FluxData *fluxdata,
					const MasterPosition &pt)
  const
{
  // The heat flux matrix M_{ij} multiplies the vector of nodal
  // temperatures to give the vector heat current J at point pt.
  // M_{ij} = epsilon_{ik} grad_k N_j
  // J_i = M_{ij} T_j
  // where N_j is the shapefunction at node j.

  if(*flux != *total_polarization) {
    throw ErrProgrammingError("Unexpected flux", __FILE__, __LINE__);
  }

  double sf = j.shapefunction(pt);
  double dsf0 = j.dshapefunction(0, pt);
  double dsf1 = j.dshapefunction(1, pt);

  SymmMatrix3 permit = permittivityTensor(mesh, el, pt);

  // Loop over flux components.  Loop over all components, even if
  // the flux is in-plane, because the out-of-plane components of
  // the flux matrix are used to construct the constraint equation.
  for(VectorFieldIterator i; !i.end(); ++i) {
    // in-plane temperature gradient contributions
    fluxdata->matrix_element(i, voltage, j) -=
      permit(i.integer(), 0) * dsf0 +
      permit(i.integer(), 1) * dsf1;
    
    // out-of-plane temperature gradient contribution
    if(!voltage->in_plane(mesh)) {
      fluxdata->matrix_element(i, voltage->out_of_plane(), j)
	-= permit(i.integer(), 2) * sf;
    }
  }
}


void DielectricPermittivity::output(const FEMesh *mesh,
				    const Element *element,
				    const PropertyOutput *output,
				    const MasterPosition &pos,
				    OutputVal *data)
  const 
{
  const std::string &outputname = output->name();
  if(outputname == "Energy") {
    // The parameter is a Python Enum instance.  Extract its name.
    std::string etype = output->getEnumParam("etype");
    if(etype == "Total" || etype == "Electric") {
      ScalarOutputVal *edata = 
	dynamic_cast<ScalarOutputVal*>(data);
      VECTOR_D E_field(3,0);
      findElectricField(mesh, element, pos, E_field);
      VECTOR_D D_field(3,0);
      SymmMatrix3 permit = permittivityTensor(mesh, element, pos);
      for(int i=0; i<3; i++)
	for(int j=0; j<3; j++)
	  D_field(i) += permit(i,j)*E_field(j);
      double e = 0;
      for(int i=0; i<3; i++) {
	e += D_field(i)*E_field(i);
      }
      *edata += 0.5*e;
    }
  }
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

IsoDielectricPermittivity::IsoDielectricPermittivity(PyObject *reg,
						     const std::string &nm,
						     double epsilon)
  : DielectricPermittivity(reg, nm),
    epsilon_(epsilon)
{
}  

void IsoDielectricPermittivity::precompute(FEMesh *mesh) {
  DielectricPermittivity::precompute(mesh);
  permittivitytensor_(0,0) = permittivitytensor_(1,1)
    = permittivitytensor_(2,2) = epsilon_;
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

AnisoDielectricPermittivity::AnisoDielectricPermittivity(PyObject *reg,
							 const std::string &nm,
							 SymmMatrix3 *k)
  : DielectricPermittivity(reg,nm),
    epsilon_(*k)
{
}

void AnisoDielectricPermittivity::cross_reference(Material *mat) {
  orientation = 
       dynamic_cast<OrientationPropBase*>(mat->fetchProperty("Orientation"));
}

void AnisoDielectricPermittivity::precompute(FEMesh *mesh) {
  DielectricPermittivity::precompute(mesh);
  Trace("AnisoDielectricPermittivity::precompute");
  if(orientation->constant_in_space())
    permittivitytensor_ = epsilon_.transform(orientation->orientation());
}

const SymmMatrix3
AnisoDielectricPermittivity::permittivityTensor(const FEMesh *mesh,
						const Element *element,
						const MasterPosition &x) const
{ 
  if(orientation->constant_in_space())
    return permittivitytensor_;
  return epsilon_.transform(orientation->orientation(mesh, element, x));
}

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

ChargeDensity::ChargeDensity(PyObject *reg, 
			       const std::string &name, double qd)
  : Property(name,reg),
    qdot_(qd) {
    total_polarization = dynamic_cast<VectorFlux*>
      (Flux::getFlux("Total_Polarization"));
}

int ChargeDensity::integration_order(const CSubProblem*, const Element*) const {
  return 0;
}

// HeatSource adds to the right-hand side of the heat equation, 
// so we want fluxrhs and not fluxmatrix.
void ChargeDensity::fluxrhs(const FEMesh *mesh, const Element *element,
			    const Flux *flux, FluxData *fluxdata,
			    const MasterPosition &masterpos) const {

  if(*flux != *total_polarization) {
    throw ErrProgrammingError("Unexpected flux.", __FILE__, __LINE__);
  }

  // In fact, there is only one component, so this is excessive.
//   for(int i=0;i<total_polarization->divergence_dim();i++) {
//     fluxdata->rhs_element(i) -= qdot_;
//   }
  fluxdata->rhs_element(0) -= qdot_;
}
