# -*- python -*-
# $RCSfile: scpatch.py,v $
# $Revision: 1.9 $
# $Author: reida $
# $Date: 2007/05/04 19:14:24 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from oof2.engine import skeletonnode

class SCPatch:
    def __init__(self, skel, femesh, cnd, mat):
        self.skel = skel
        self.femesh = femesh
        self.assembly_node = cnd
        self.mat = mat  # material
        self.elements = []  # Skeleton elements
        self.segments = {}  # Skeleton segments, stored to be looked at later
        self.recovery_nodes = {}  # FEMesh nodes for recovery

    def getMaterial(self):
        return self.mat

    def nElements(self):
        return len(self.elements)

    def getElements(self):  # returns a copy of self.elements
        return self.elements[:]

    def addElement(self, el):           # el is a Skeleton element
        self.elements.append(el)
        snd0 = el.nodes[-1]  # Skeleton node
        for snd1 in el.nodes:
            # If the edge is on boundary (external or internal),
            # all the nodes on the edge are recovery nodes.
            if (self.femesh.bdy_node_map[snd0.meshindex] and
                           self.femesh.bdy_node_map[snd1.meshindex]):
                self.recovery_nodes[snd0.meshindex] = 1
                self.recovery_nodes[snd1.meshindex] = 1
                if self.femesh.order() > 1:  # higher order femesh
                    edge_nodes = self.femesh.getEdgeNodes(
                        skeletonnode.canonical_order(snd0,snd1))
                    for fnd in edge_nodes:
                        self.recovery_nodes[fnd.index()] = 1
            else:
                # Non-boundary edges(segments) need to be looked at later.
                self.segments[self.skel.findSegment(snd0, snd1)] = 1

            snd0 = snd1  # for the next iteration

    def findRecoveryNodes(self, qualified=1):
        # nodes on patch boundary but not on actual boundary.
        patch_boundary = {}
        for seg in self.segments.keys():
            elcount = 0
            for el in seg.getElements():
                if el in self.elements:
                    elcount += 1
            if elcount == 1:  # the segment is on patch boundary
                for nd in seg.nodes():
                    if not self.femesh.bdy_node_map[nd.meshindex]:
                        patch_boundary[nd] = 1
                self.segments[seg] = 0  # no need to look at it again.

        # recovery nodes that are inside patch (except internal nodes).
        for seg, need2look in self.segments.items():
            if not need2look:
                continue
            seg_nodes = seg.nodes()
            for nd in seg_nodes:
                if not patch_boundary.has_key(nd):
                    self.recovery_nodes[nd.meshindex] = 1
            if self.femesh.order() > 1:  # higher order femesh
                edge_nodes = self.femesh.getEdgeNodes(seg_nodes)
                for nd in edge_nodes:
                    self.recovery_nodes[nd.index()] = 1

        # Add any internal nodes
        if self.femesh.internal_nodes:
            for el in self.elements:
                for nd in self.femesh.getInternalNodes(el):
                    self.recovery_nodes[nd.index()] = 1

        # Creating a C++ CSCPatch object that will do actual computation.
        self.femesh.add_scpatch(self.assembly_node.meshindex, self.mat,
                                self.femesh.order(),
                                [el.meshindex for el in self.elements],
                                self.recovery_nodes.keys(),
                                qualified)
        
    def __repr__(self):
        return `[el.meshindex for el in self.elements]`

#########################################

class SCPatchCollection:
    def __init__(self, femesh):
        self.femesh = femesh
        self.patches = {}  # keyed by (node, material) pair
        ## TODO: Why not call initialize() and finalize() directly
        ## from __init__()?
    def buildPatches(self, skel):
        self.initialize(skel)
        self.finalize()

    def initialize(self, skel):
        # pre-processing -- finding nodes on boundaries (material & geometric).
        # it helps find recovery nodes more efficiently.
        self.skel = skel
        self.femesh.create_bdy_node_map(skel)
            
        for cnd in skel.node_iterator():  # Skeleton node!
            # Among neighbors, elements will be grouped according to their
            # respective materials.
            for el in cnd.neighborElements(skel):
                self.addElementToPatch(skel, cnd, el)

    def addElementToPatch(self, skel, cnd, el):  # Skeleton element
        mat = self.femesh.getElement(el.meshindex).material()
        # Do not add elements or create patches where no material exists.
        if mat:
            try:  # existing entity
                self.patches[(cnd, mat)].addElement(el)
            except KeyError:  # new entity
                self.patches[(cnd, mat)] = SCPatch(skel, self.femesh, cnd, mat)
                self.patches[(cnd, mat)].addElement(el)

    def finalize(self):
        # Augment patches and assign recovery nodes to patches.
        # For linear femesh, at least 3 elements are needed to
        # determine coefficients for [1, x, y]
        # For quadratic femesh, 2 elements are needed -- [1,x,y,x,y,x^2,y^2]
        if self.femesh.order() == 1:  # Linear
            min_els = 4
        else:  # Quadratic
            min_els = 4  # 4 seems to do better than 2
        for patch in self.patches.values():
            if patch.nElements() < min_els:
                # Collect every available neighbor element from
                # all the nodes in the current patch.
                for el in patch.getElements():
                    for nd in el.nodes:
                        for el in nd.neighborElements(self.skel):
                            fel = self.femesh.getElement(el.meshindex)
                            fel_mat = fel.material()
                            if fel_mat == patch.getMaterial() and el not in patch.elements:
                                patch.addElement(el)

            # If still not enough elements, then it's a disqualified patch.
            if patch.nElements() < min_els:
                patch.findRecoveryNodes(qualified=0)
            else:
                patch.findRecoveryNodes()
