# -*- python -*-
# $RCSfile: snaprefine.py,v $
# $Revision: 1.14 $
# $Author: rlua $
# $Date: 2007/03/27 20:02:23 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file implements "Snap Refine"
# (was called "Intelligent Refine" (pun intended)).
# New edge nodes are transition points.
# The "ruleset" is always liberal.

# Comment from the original refine.py
# Refinement works in two stages.  First a RefinementTarget object (from
# refinementtarget.py) checks the elements and marks edges for
# subdivision.  Then a RefinementRuleSet (from refinemethod.py) is
# used to actually refine the elements, according to how many times
# each edge is marked.

from oof2.SWIG.common.IO import progressbar
from oof2.common import debug
from oof2.common import enum
from oof2.common import parallel_enable
from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import refinementtarget
from oof2.engine import refinequadbisection
from oof2.engine import skeletonmodifier
from oof2.engine import skeletonsegment
from oof2.engine import skeletonnode

from oof2.engine import refine
from oof2.engine import snaprefinemethod

SkeletonSegment = skeletonsegment.SkeletonSegment

arbitrary_factor = 128          # used by findSignature().  Must be
				# larger than the largest number of
				# refinements of an edge.

#################################

class EdgeMarkings:
    # Class for storing and returning how many divisions should be
    # performed on an edge.  Edges are defined by a pair of nodes.
    def __init__(self, newSkeleton, oldSkeleton, maxdelta2):
        self.markings = {}
        # New for Snap Refine
        self.newEdgeNodes = {}
        self.newSkeleton=newSkeleton
        self.newSkelMS=newSkeleton.MS
        self.oldSkeleton=oldSkeleton
        self.maxdelta2=maxdelta2
        
    def mark(self, node0, node1, ndivs):
        # arguments are the nodes defining the edge, and the number of
        # new nodes to add to that edge.
        key = skeletonnode.canonical_order(node0, node1)
        try:
            # Only mark an edge if it's not already marked for more
            # divisions.
            if self.markings[key] < ndivs:
                self.markings[key] = ndivs
        except KeyError:
            self.markings[key] = ndivs

        # mark the partner segment, if it exists
        partners = node0.getPartnerPair(node1)
        if partners is not None:
            partnerKey = skeletonnode.canonical_order(partners[0], partners[1])
            try:
                # Only mark an edge if it's not already marked for more
                # divisions.
                if self.markings[partnerKey] < ndivs:
                    self.markings[partnerKey] = ndivs
            except KeyError:
                self.markings[partnerKey] = ndivs

    def getMark(self, node0, node1):
        key = skeletonnode.canonical_order(node0, node1)
        try:
            return self.markings[key]
        except KeyError:
            return 0

    def getMarks(self, element):
        nnodes = element.nnodes()
        return [self.getMark(element.nodes[i], element.nodes[(i+1)%nnodes])
                for i in range(nnodes)]

    def getTransitionPoints(self, n0pt, n1pt):
        #Get a transition point starting from each end.
        transpt0 = self.newSkelMS.transitionPointWithPoints_unbiased(n0pt,n1pt)
        transpt1 = self.newSkelMS.transitionPointWithPoints_unbiased(n1pt,n0pt)
        if transpt0 and transpt1:
            #Check if the transition points are identical or too close together
            #Steve doesn't like this, unless self.maxdelta2==0.
            #He might prefer this instead: if transpt0 != transpt1: ...
            if (transpt0-transpt1)**2>self.maxdelta2:
                transptlist=[]
                if (transpt0-n0pt)**2>self.maxdelta2:
                    transptlist.append(transpt0)
                if (transpt1-n1pt)**2>self.maxdelta2:
                    transptlist.append(transpt1)
                return transptlist
            else:
                #Pick one. Check if it is at, or too close to, the end points.
                if (transpt0-n0pt)**2>self.maxdelta2 and \
                   (transpt0-n1pt)**2>self.maxdelta2:
                    return [transpt0]
                else:
                    return []
        elif transpt0:
            if (transpt0-n0pt)**2>self.maxdelta2 and \
                   (transpt0-n1pt)**2>self.maxdelta2:
                return [transpt0]
            else:
                return []
        elif transpt1:
            if (transpt1-n0pt)**2>self.maxdelta2 and \
                   (transpt1-n1pt)**2>self.maxdelta2:
                return [transpt1]
            else:
                return []
        return []

    # Modified from refine.py to take into account the
    # edge categories and transition points.
    # This method does the work of both getMarks and getNewEdgeNodes
    # This method selects which marked edges are to be bisected.
    def getSnapMarks(self, element):
        numinitmarks=0
        intelmarks=[]
        intelcats=[]
        inteltranspts=[]
        nnodes = element.nnodes()
        for i in range(nnodes):
            node0=element.nodes[i]
            node1=element.nodes[(i+1)%nnodes]
            n0pt=node0.position()
            n1pt=node1.position()
            key=skeletonnode.canonical_order(node0,node1)
            initmark=0
            try:
                initmark=self.markings[key]
                #The fact that the edge has an entry in the
                #markings dictionary is a sign that the user
                #intends for the element to be refined.
                #If numinitmarks is non-zero, then the element will be
                #refined in some fashion, even if the edges are not
                #bisected (i.e. no transition points found).
                numinitmarks+=1
            except KeyError:
                initmark=-1
            if initmark>0:
                #Marked for bisection, find or retrieve new edge nodes.
                transptlist=[]
                transptnodelist=[]
                try:
                    #See if we already created new edge nodes for this edge.
                    transptnodelist = self.newEdgeNodes[key][:] #Make a list copy
                    transptnodelist.reverse()
                    #Get the list of transition points.
                    for transptnode in transptnodelist:
                        transptlist.append(transptnode.position())
                    #transptlist.reverse()
                except KeyError:
                    partners=node0.getPartnerPair(node1)
                    #It can happen that partners contains the same two nodes (node1,node0)!
                    if partners and node0!=partners[1]:
                        ############## Begin Periodic Skeleton Node Construction ###################
                        partnerKey = skeletonnode.canonical_order(partners[0], partners[1])
                        #If the edge nodes have periodic partners, then find the transition
                        #points on the current edge and the partner edge. Get the pair of
                        #transition points that is closest to the endpoints (imagine joining
                        #together the two opposite edges).
                        n0pt_partner=partners[0].position()
                        n1pt_partner=partners[1].position()
                        transptlistCandidates=self.getTransitionPoints(n0pt,n1pt)+\
                                         self.getTransitionPoints(n0pt_partner,n1pt_partner)
                        sx=self.newSkelMS.size()[0]
                        sy=self.newSkelMS.size()[1]
                        if len(transptlistCandidates)>0:
                            #Case 1: vertical boundary
                            if n0pt.x==0 and n1pt.x==0:
                                #Get transpt closest to the endpoints
                                transpt0=transptlistCandidates[0]
                                transpt1=transptlistCandidates[0]
                                #Squaring would not be necessary if we order
                                #the operands correctly.
                                min0=(transpt0.y-n0pt.y)**2
                                min1=(transpt1.y-n1pt.y)**2
                                for transpt in transptlistCandidates:
                                    mintmp=(transpt.y-n0pt.y)**2
                                    if mintmp<min0:
                                        min0=mintmp
                                        transpt0=transpt
                                    mintmp=(transpt.y-n1pt.y)**2
                                    if mintmp<min1:
                                        min1=mintmp
                                        transpt1=transpt
                                transptlist.append(transpt0)
                                #The second boolean condition may suffice.
                                if transpt0 != transpt1 and \
                                       (transpt0.y-transpt1.y)**2>self.maxdelta2:
                                    transptlist.append(transpt1)
                                #Fix the horizontal coordinates to match the location
                                #of the edge when creating nodes.
                                transptnodelistPartner=[]
                                for transpt in transptlist:
                                    #For current edge
                                    transptnode = self.newSkeleton.newNode(0,transpt.y)
                                    transptnodelist.append(transptnode)
                                    #For partner
                                    transptnodePartner = self.newSkeleton.newNode(sx,transpt.y)
                                    transptnode.addPartner(transptnodePartner)
                                    transptnodelistPartner.append(transptnodePartner)
                                self.newEdgeNodes[key] = transptnodelist
                                #transptnodelistPartner.reverse()
                                self.newEdgeNodes[partnerKey] = transptnodelistPartner
                            #Case 2: vertical boundary
                            elif n0pt.x==sx and n1pt.x==sx: #vertical boundary
                                #Get transpt closest to the endpoints
                                transpt0=transptlistCandidates[0]
                                transpt1=transptlistCandidates[0]
                                #Squaring would not be necessary if we order
                                #the operands correctly.
                                min0=(transpt0.y-n0pt.y)**2
                                min1=(transpt1.y-n1pt.y)**2
                                for transpt in transptlistCandidates:
                                    mintmp=(transpt.y-n0pt.y)**2
                                    if mintmp<min0:
                                        min0=mintmp
                                        transpt0=transpt
                                    mintmp=(transpt.y-n1pt.y)**2
                                    if mintmp<min1:
                                        min1=mintmp
                                        transpt1=transpt
                                transptlist.append(transpt0)
                                if transpt0 != transpt1 and \
                                       (transpt0.y-transpt1.y)**2>self.maxdelta2:
                                    transptlist.append(transpt1)
                                #Fix the horizontal coordinates to match the location
                                #of the edge when creating nodes.
                                transptnodelistPartner=[]
                                for transpt in transptlist:
                                    #For current edge
                                    transptnode = self.newSkeleton.newNode(sx,transpt.y)
                                    transptnodelist.append(transptnode)
                                    #For partner
                                    transptnodePartner = self.newSkeleton.newNode(0,transpt.y)
                                    transptnode.addPartner(transptnodePartner)
                                    transptnodelistPartner.append(transptnodePartner)
                                self.newEdgeNodes[key] = transptnodelist
                                #transptnodelistPartner.reverse()
                                self.newEdgeNodes[partnerKey] = transptnodelistPartner
                            #Case 3: horizontal boundary
                            elif n0pt.y==0 and n1pt.y==0:
                                #Get transpt closest to the endpoints
                                transpt0=transptlistCandidates[0]
                                transpt1=transptlistCandidates[0]
                                #Squaring would not be necessary if we order
                                #the operands correctly.
                                min0=(transpt0.x-n0pt.x)**2
                                min1=(transpt1.x-n1pt.x)**2
                                for transpt in transptlistCandidates:
                                    mintmp=(transpt.x-n0pt.x)**2
                                    if mintmp<min0:
                                        min0=mintmp
                                        transpt0=transpt
                                    mintmp=(transpt.x-n1pt.x)**2
                                    if mintmp<min1:
                                        min1=mintmp
                                        transpt1=transpt
                                transptlist.append(transpt0)
                                if transpt0 != transpt1 and \
                                       (transpt0.x-transpt1.x)**2>self.maxdelta2:
                                    transptlist.append(transpt1)
                                #Fix the horizontal coordinates to match the location
                                #of the edge when creating nodes.
                                transptnodelistPartner=[]
                                for transpt in transptlist:
                                    #For current edge
                                    transptnode = self.newSkeleton.newNode(transpt.x,0)
                                    transptnodelist.append(transptnode)
                                    #For partner
                                    transptnodePartner = self.newSkeleton.newNode(transpt.x,sy)
                                    transptnode.addPartner(transptnodePartner)
                                    transptnodelistPartner.append(transptnodePartner)
                                self.newEdgeNodes[key] = transptnodelist
                                #transptnodelistPartner.reverse()
                                self.newEdgeNodes[partnerKey] = transptnodelistPartner
                            #Case 4: horizontal boundary
                            #elif n0pt.y==sy and n1pt.y==sy:
                            else:
                                #Get transpt closest to the endpoints
                                transpt0=transptlistCandidates[0]
                                transpt1=transptlistCandidates[0]
                                #Squaring would not be necessary if we order
                                #the operands correctly.
                                min0=(transpt0.x-n0pt.x)**2
                                min1=(transpt1.x-n1pt.x)**2
                                for transpt in transptlistCandidates:
                                    mintmp=(transpt.x-n0pt.x)**2
                                    if mintmp<min0:
                                        min0=mintmp
                                        transpt0=transpt
                                    mintmp=(transpt.x-n1pt.x)**2
                                    if mintmp<min1:
                                        min1=mintmp
                                        transpt1=transpt
                                transptlist.append(transpt0)
                                if transpt0 != transpt1 and \
                                       (transpt0.x-transpt1.x)**2>self.maxdelta2:
                                    transptlist.append(transpt1)
                                #Fix the horizontal coordinates to match the location
                                #of the edge when creating nodes.
                                transptnodelistPartner=[]
                                for transpt in transptlist:
                                    #For current edge
                                    transptnode = self.newSkeleton.newNode(transpt.x,sy)
                                    transptnodelist.append(transptnode)
                                    #For partner
                                    transptnodePartner = self.newSkeleton.newNode(transpt.x,0)
                                    transptnode.addPartner(transptnodePartner)
                                    transptnodelistPartner.append(transptnodePartner)
                                self.newEdgeNodes[key] = transptnodelist
                                #transptnodelistPartner.reverse()
                                self.newEdgeNodes[partnerKey] = transptnodelistPartner
                        else:
                            self.newEdgeNodes[key]=[]
                            self.newEdgeNodes[partnerKey]=[]
                        ############## End Periodic Skeleton Node Construction ###################
                    else:
                        transptlist=self.getTransitionPoints(n0pt,n1pt)
                        #Not retrieved from newEdgeNodes. Create new skeleton node(s) and
                        #add entry to newEdgeNodes dict.
                        for transpt in transptlist:
                            transptnode = self.newSkeleton.newNode(transpt.x,transpt.y)
                            transptnodelist.append(transptnode)
                        self.newEdgeNodes[key] = transptnodelist

                #transptlist is a list of transition points (at most 2),
                #either just created with getTransitionPoints,
                #or retrieved from newEdgeNodes (put in there when a
                #neighboring element was visited).
                if len(transptlist)==2:
                    transpt0=transptlist[0]
                    transpt1=transptlist[1]
                    #Order of the arguments to edgeHomogeneityCat are significant.
                    #If the edge n0pt-n1pt is horizontal or vertical and lies at
                    #a pixel boundary, then we want the function to use the pixels
                    #lying within the element in its calculations.
                    homog0, cat0 = self.newSkelMS.edgeHomogeneityCat(n0pt,transpt0)
                    homog1, cat1 = self.newSkelMS.edgeHomogeneityCat(transpt0,transpt1)
                    homog2, cat2 = self.newSkelMS.edgeHomogeneityCat(transpt1,n1pt)
                    intelmarks.append(2)
                    intelcats.append((cat0,cat1,cat2))
                    inteltranspts.append(transptnodelist)
                elif len(transptlist)==1:
                    transpt=transptlist[0]
                    #Order of the arguments to edgeHomogeneityCat are significant.
                    #If the edge n0pt-n1pt is horizontal or vertical and lies at
                    #a pixel boundary, then we want the function to use the pixels
                    #lying within the element in its calculations.
                    homog0, cat0 = self.newSkelMS.edgeHomogeneityCat(n0pt,transpt)
                    homog1, cat1 = self.newSkelMS.edgeHomogeneityCat(transpt,n1pt)
                    intelmarks.append(1)
                    intelcats.append((cat0,cat1))
                    inteltranspts.append(transptnodelist)
                else:
                    intelmarks.append(0)
                    homogedge,catedge=self.newSkelMS.edgeHomogeneityCat(n0pt,n1pt)
                    intelcats.append((catedge,))
                    inteltranspts.append([])
            else:
                if initmark>-1: #initially marked but was not bisected
                    #If it gets to this point, initmark must have been zero.
                    #This should not happen.
                    self.newEdgeNodes[key]=[]
                intelmarks.append(0)
                homogedge,catedge=self.newSkelMS.edgeHomogeneityCat(n0pt,n1pt)
                intelcats.append((catedge,))
                inteltranspts.append([])

        return (numinitmarks, intelmarks, intelcats, inteltranspts)

    def getNMarkedEdges(self, element):
        marks = self.getMarks(element)
        nmarks = 0
        for m in marks:
            if m: nmarks += 1
        return nmarks

###################################
            
class SnapRefine(skeletonmodifier.SkeletonModifier):
    #Unlike refine.py, alpha parameter not used (yet?)
    def __init__(self, targets, criterion, min_distance, alpha=1):
        self.targets = targets      # RefinementTarget instance
        self.criterion = criterion  # Criterion for refinement
        # alpha is used for deciding between different possible
        # refinements, using effective energy
        self.alpha = alpha 
        # Available rules
        self.rules = snaprefinemethod.getRuleSet('liberal')

        self.min_distance=min_distance
    #########

    def applyAMR(self, skeleton, context): # for adaptive mesh refinement
        newSkeleton = skeleton.improperCopy(fresh=True)
        return self.refinement(skeleton, newSkeleton, context)
    
    def apply(self, skeleton, context):
        newSkeleton = skeleton.improperCopy(skeletonpath=context.path())  # fresh=False
        return self.refinement(skeleton, newSkeleton, context)

    def refinement(self, skeleton, newSkeleton, context):
        # pBar.set_message("marking elements...       ")
        pBar = progressbar.getProgress()

        (maxdelta,deltay)=newSkeleton.MS.sizeOfPixels()
        if maxdelta<deltay:
            maxdelta=deltay
        maxdelta2=(self.min_distance*maxdelta)**2
        markedEdges = EdgeMarkings(newSkeleton,skeleton,maxdelta2)
        self.newEdgeNodes = {}          # allows sharing of new edge nodes

        # Primary marking (bisections only!)
        self.targets(skeleton, context, 1, markedEdges,
                     self.criterion)

        # Additional marking
        #self.degree.markExtras(skeleton, markedEdges)

        # Refine elements and segments
        segmentdict = {}                # which segments have been handled
        elements = skeleton.elements
        n = len(elements)
        
        for ii in range(n):
            oldElement = elements[ii]
            oldnnodes = oldElement.nnodes()
            # Get list of number of subdivisions on each edge ("marks")
            (numinitmarks,marks,cats,edgenodes) = \
                          markedEdges.getSnapMarks(oldElement)

            # Find the canonical order for the marks. (The order is
            # ambiguous due to the arbitrary choice of the starting
            # edge.  Finding the canonical order allows the refinement
            # rule to be found in the rule table.)  rotation is the
            # offset into the elements node list required to match the
            # refinement rule to the element's marked edges.
            # signature is the canonical ordering of the marks.
            rotation, signature = findSignature(marks)
            
            # Create new elements
            if numinitmarks>0:
                newElements = self.rules[signature].apply(oldElement, rotation,
                                                          cats,
                                                          edgenodes, newSkeleton,
                                                          maxdelta2)
            else:
                newElements = snaprefinemethod.unrefinedelement(oldElement,
                                                                 rotation,
                                                                 newSkeleton)
            # If the old element's homogeneity is 1, it's safe to say that
            # new elements' homogeneities are 1.
            if oldElement.homogeneity(skeleton.MS) == 1.0:
                for el in newElements:
                    el.copyHomogeneity(oldElement)

            # The calls to Skeleton.newElement() made by the
            # refinement rules have created new SkeletonSegments in
            # newSkeleton, but have not set the parentage of those
            # segments.  We have to fix that here.
            for newElement in newElements:
                for segment in newElement.getSegments(newSkeleton):
                    # Only look at each segment once.
                    if not segmentdict.has_key(segment):
                        segmentdict[segment] = 1
                        pseg = findParentSegment(skeleton, newElement,
                                                 segment,
                                                 edgenodes)
                        if pseg:
                            pseg.add_child(segment)
                            segment.add_parent(pseg)
            if pBar.query_stop():
                pBar.set_failure()
                pBar.set_message("Failed")
                return None
            else:
                pBar.set_progress(1.0*(ii+1)/n)
                pBar.set_message("refining skeleton: %d/%d" % (ii+1, n))
       
        newSkeleton.cleanUp()

        return newSkeleton

#################################

def findParentSegment(oldSkeleton, element, segment, edgenodes):
    n0, n1 = segment.nodes()            # nodes of the segment
    parents = n0.getParents()
    if parents:
        p0 = parents[-1]
    else:
        p0 = None
        
    parents = n1.getParents()
    if parents:
        p1 = parents[-1]
    else:
        p1 = None    
    oldElement = element.getParents()[-1]
    nnodes = oldElement.nnodes()

    if p0 is not None and p1 is not None:
        pseg = oldSkeleton.findSegment(p0, p1)
        if pseg:
            # Trivial case: the parent nodes define a segment in the
            # parent skeleton.  That segment is this segment's parent.
            return pseg
        # Both nodes have parents, but the parent nodes don't define a
        # segment in the old skeleton.  (This is unlikely to happen
        # during refinement.)  The segment does not have a parent.
        return None

    elif p0 is not None or p1 is not None:
        # Exactly one new endpoint has a parent in the old element.
        # The other endpoint may lie on an old segment.  If that
        # segment's other node is the parent of the new node, then
        # that segment is the parent of this segment.
        if p0 is None:
            parentnode = p1
            childnode = n1
            freechild = n0
        else:
            parentnode = p0
            childnode = n0
            freechild = n1
        which = oldElement.nodes.index(parentnode)
        if freechild in edgenodes[which]:
            next = (which+1)%nnodes
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[next])
        prev = (which - 1 + nnodes) % nnodes
        if freechild in edgenodes[prev]:
            return oldSkeleton.findSegment(parentnode, oldElement.nodes[prev])

    elif p0 is None and p1 is None:
        # Neither new endpoint has a parent.  If the new endpoints are
        # consecutive nodes in the list of new nodes added to a parent
        # element segment, then that parent segment is the parent
        # we're looking for.
        for i in range(nnodes):
            nodelist = edgenodes[i]
            try:
                which0 = nodelist.index(n0)
                which1 = nodelist.index(n1)
            except ValueError:
                pass
            else:
                if abs(which0 - which1) == 1:
                    return oldSkeleton.findSegment(oldElement.nodes[i],
                                                   oldElement.nodes[(i+1)%
                                                                    nnodes])

################################

def findSignature(marks):
    # Given a list of subdivisions of sides (marks), rotate a
    # canonical starting point for the list, so that it can be used as
    # an index into a table of refinement functions.  For a list
    # [x,y,z], the canonical order is that which maximizes the number
    # xyz in base arbitrary_factor.  This will fail if edges are ever
    # divided into more than arbitrary_factor segments in one
    # refinement operation.
    n = len(marks)
    max = -1
    imax = None
    for i in range(n):
        key = marks[i]
        for j in range(1, n):
            key = arbitrary_factor*key + marks[(i+j)%n]
        if key > max:
            max = key
            imax = i
    return imax, tuple([marks[(i+imax)%n] for i in range(n)])


####################

registeredclass.Registration(
    'Snap Refine',
    skeletonmodifier.SkeletonModifier,
    SnapRefine,
    ordering=400,
    params=[parameter.RegisteredParameter('targets',
                                          refinementtarget.RefinementTarget,
                                          tip='Target elements to be refined.'),
            parameter.RegisteredParameter('criterion', refine.RefinementCriterion,
                                          tip='Exclude certain elements.'),
            parameter.FloatParameter('min_distance', value=1.0,
                                     tip="""Minimum distance of transition points, along a given edge,
from each other and the end-points, in units of the maximum pixel dimension.""")
#            ,skeletonmodifier.alphaParameter
            ],
    tip="Subdivide elements along pixel boundaries.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/snaprefine.xml')
    )
