// -*- C++ -*-
// $RCSfile: steppingscheme.h,v $
// $Revision: 1.25 $
// $Author: langer $
// $Date: 2006/12/19 20:29:44 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>
#ifndef STEPPINGSCHEME_H
#define STEPPINGSCHEME_H

class CSRmatrix;
class CSubProblem;
class LinearSolver;
class SolverDriverCore;
#include "engine/solverdriver.h"



class SteppingScheme {
private:
  double dt;
  double tf;
  VECTOR_D *x_copy;
  //OOF 'standard' time is held by mesh
public:
  SteppingScheme() : dt(0), tf(0), x_copy(0) {}
  SteppingScheme(const double dt, const double tf)
    : dt(dt), tf(tf), x_copy(0) {}
  virtual ~SteppingScheme();
  //
  VECTOR_D * get_x() {return x_copy;}
  void set_x(VECTOR_D *x);//allocates memory!!!
  //
  void set_dt(const double deetee) {dt = deetee;}
  double get_dt() const {return dt;}
  void set_tf(const double teef) {tf = teef;}
  double get_tf() {return tf;}

  //hooks for subclasses
  virtual void precompute(CSubProblem *,
			  CSRmatrix *A, VECTOR_D *rhs, VECTOR_D *x) = 0;
  virtual CSRmatrix *get_K(CSubProblem*, CSRmatrix *A)  = 0;
  virtual VECTOR_D *get_b(CSubProblem*, CSRmatrix *A, VECTOR_D *rhs) = 0;
  virtual void clean_up() = 0;
  virtual void postcompute(CSubProblem*) = 0; // synchronizes target time
  virtual void update_mesh(CSubProblem&) {}
};


class NullScheme : public SteppingScheme {
public:
  NullScheme() : SteppingScheme() {}
  NullScheme(const double dt, const double tf) : SteppingScheme(dt, tf) {}
  virtual ~NullScheme() {}

  virtual void precompute(CSubProblem*,
			  CSRmatrix *A, VECTOR_D *rhs, VECTOR_D *x)
  {}
  //get_K and get_b do nothing. They return A and b pointers as-they-are
  //i.e., the behavior in Null
  virtual CSRmatrix * get_K(CSubProblem*, CSRmatrix *A) { return A; }
  virtual VECTOR_D * get_b(CSubProblem*, CSRmatrix *A, VECTOR_D *rhs) {
    return rhs;
  }
  virtual void clean_up() {}
  virtual void postcompute(CSubProblem*) {}
};




//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//PDEs time dependent steppers machinery starts here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


class TimeDependent : public SteppingScheme {
public:
  TimeDependent() : SteppingScheme() {}
  TimeDependent(const double dt, const double tf) : SteppingScheme(dt, tf) {}
  virtual ~TimeDependent() {}
  virtual void precompute(CSubProblem*,
			  CSRmatrix *A, VECTOR_D *rhs, VECTOR_D *x) = 0;
  virtual CSRmatrix *get_K(CSubProblem *, CSRmatrix *A) = 0;
  virtual VECTOR_D *get_b(CSubProblem *, CSRmatrix *A, VECTOR_D *rhs) = 0;
  virtual void clean_up() = 0;
  virtual void postcompute(CSubProblem*) = 0;
};



//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//quasistatic machinery starts here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
class QuasiStatic : public TimeDependent {
public:
  QuasiStatic() : TimeDependent() {}
  virtual ~QuasiStatic() {}
QuasiStatic(const double dt, const double tf) : TimeDependent(dt, tf) {}

  
  //get_K and get_b do nothing. They return A and b pointers as-they-are
  //i.e., the behavior in Null
  virtual void precompute(CSubProblem*,
			  CSRmatrix *A, VECTOR_D *rhs, VECTOR_D *x)
  {}
  virtual CSRmatrix * get_K(CSubProblem*, CSRmatrix *A) { return A; }
  virtual VECTOR_D * get_b(CSubProblem*, CSRmatrix *A, VECTOR_D *rhs) {
    return rhs;
  }
  virtual void clean_up() {}
  virtual void postcompute(CSubProblem*) {};
  virtual void update_mesh(CSubProblem&); // updates time index in mesh
};


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//quasistatic machinery ends here
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//



class Kinetic : public  TimeDependent {
private:
  const double theta;
  VECTOR_D * xn;//solution from previous step
  VECTOR_D * rhsn; //rhs from previous time step
  CSRmatrix * conductivityMatrix;

  // Storage for retaining ownership of K and b things built by get_K
  // and get_b routines, and cleared by clean_up.
  CSRmatrix *_K; 
  VECTOR_D *_b;
  
  //This class sets-up for its solution the system K x  = b
  //with
  //K = ConducivityMatrix + theta*dt*A
  //and
  //b = (ConducivityMatrix-(1.0-theta)*dt*A)*xn + dt*f_hat
  //f_hat = rhs + theta*(rhs-rhsn)
  
public:
  Kinetic(double dt, double tf, double theta)
    : TimeDependent(dt, tf),
      theta(theta),
      xn(0),
      rhsn(0),
      conductivityMatrix(0),
      _K(0),
      _b(0)
  {}
  virtual ~Kinetic();

  virtual void precompute(CSubProblem*,
			  CSRmatrix *A, VECTOR_D *rhs, VECTOR_D *x);
  virtual CSRmatrix * get_K(CSubProblem*, CSRmatrix *A);
  virtual VECTOR_D * get_b(CSubProblem*, CSRmatrix *A, VECTOR_D *rhs);
  virtual void clean_up();
  virtual void postcompute(CSubProblem*);
  virtual void update_mesh(CSubProblem&); // updates time index in mesh
};


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//
//DEBUGGING SECTION STARTS HERE
//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//


void sum_matrices();

#endif // STEPPINGSCHEME_H
